"""
Prints structures in a manner similar to WinDbg's "dt" command.
"""

from __future__ import annotations

import pwndbg
import pwndbg.aglib.memory
import pwndbg.aglib.typeinfo
import pwndbg.dbg_mod


def _field_to_human(
    f: pwndbg.dbg_mod.TypeField | pwndbg.dbg_mod.Value | pwndbg.dbg_mod.Type,
) -> str:
    if isinstance(f, pwndbg.dbg_mod.TypeField):
        t = f.type
    elif isinstance(f, pwndbg.dbg_mod.Type):
        t = f
    elif isinstance(f, pwndbg.dbg_mod.Value):
        t = f.type
    else:
        raise NotImplementedError("unknown type")

    return t.name_to_human_readable


def dt(
    name: str = "",
    addr: int | pwndbg.dbg_mod.Value | None = None,
    obj: pwndbg.dbg_mod.Value | None = None,
) -> str:
    """
    Dump out a structure type WinDbg style.
    """
    # Return value is a list of strings.of
    # We concatenate at the end.
    rv: list[str] = []

    if obj and not name:
        t = obj.type
        while t.code == pwndbg.dbg_mod.TypeCode.POINTER:
            t = t.target()
            obj = obj.dereference()
        name = str(t)

    # Lookup the type name specified by the user
    else:
        t = pwndbg.aglib.typeinfo.load(name)

    if not t:
        return "Type not found."

    # If it's not a struct (e.g. int or char*), bail
    if t.strip_typedefs().code not in (
        pwndbg.dbg_mod.TypeCode.STRUCT,
        pwndbg.dbg_mod.TypeCode.UNION,
    ):
        return f"Not a structure: {t.strip_typedefs().name_to_human_readable}"

    # If an address was specified, create a Value of the
    # specified type at that address.
    if addr is not None:
        obj = pwndbg.aglib.memory.get_typed_pointer_value(t, addr)

    # Header, optionally include the name
    header = name
    if obj:
        header = f"{header} @ {hex(int(obj.address))}"
    rv.append(header)

    iter_fields = [(field.name, field) for field in t.fields()]

    for field_name, field in iter_fields:
        # Offset into the parent structure
        offset = field.bitpos // 8
        bitpos = field.bitpos % 8
        ftype = field.type.strip_typedefs()
        extra = _field_to_human(field)

        if obj and obj.type.strip_typedefs().code in (
            pwndbg.dbg_mod.TypeCode.STRUCT,
            pwndbg.dbg_mod.TypeCode.UNION,
        ):
            try:
                obj_value = obj[field_name]
                if ftype.code == pwndbg.dbg_mod.TypeCode.INT:
                    extra = hex(int(obj_value))
                elif (
                    ftype.code in (pwndbg.dbg_mod.TypeCode.POINTER, pwndbg.dbg_mod.TypeCode.ARRAY)
                    and ftype.target() == pwndbg.aglib.typeinfo.uchar
                ):
                    data = pwndbg.aglib.memory.read(int(obj_value.address), ftype.sizeof)
                    extra = " ".join(f"{b:02x}" for b in data)
                else:
                    extra = obj_value.value_to_human_readable()
            except pwndbg.dbg_mod.Error as e:
                return f"{e}\nIs the provided address near a page boundry?"

        # Adjust trailing lines in 'extra' to line up
        # This is necessary when there are nested structures.
        # Ideally we'd expand recursively if the type is complex.
        extra_lines: list[str] = []
        for i, line in enumerate(str(extra).splitlines()):
            if i == 0:
                extra_lines.append(line)
            else:
                extra_lines.append(35 * " " + line)
        extra = "\n".join(extra_lines)

        bitpos_str = "" if not bitpos else (f".{bitpos}")

        if obj:
            line = f"    0x{int(obj.address) + offset:016x} +0x{offset:04x}{bitpos_str} {field_name:<20} : {extra}"
        else:
            line = f"    +0x{offset:04x}{bitpos_str} {field_name:<20} : {extra}"
        rv.append(line)

    return "\n".join(rv)



from __future__ import annotations

import argparse

import pwndbg.aglib.file
import pwndbg.aglib.proc
import pwndbg.aglib.qemu
import pwndbg.commands
from pwndbg.color import message
from pwndbg.commands import CommandCategory

if pwndbg.dbg.is_gdblib_available():
    import gdb


def check_aslr() -> tuple[bool | None, str]:
    """
    Detects the ASLR status. Returns True, False or None.

    None is returned when we can't detect ASLR.
    """
    # QEMU does not support this concept.
    if pwndbg.aglib.qemu.is_qemu():
        return None, "Could not detect ASLR on QEMU targets"

    # Systemwide ASLR is disabled
    try:
        data = pwndbg.aglib.file.get("/proc/sys/kernel/randomize_va_space")
        if b"0" in data:
            return False, "kernel.randomize_va_space == 0"
    except Exception:
        print("Could not check ASLR: can't read randomize_va_space")

    # Check the personality of the process
    if pwndbg.aglib.proc.alive():
        try:
            data = pwndbg.aglib.file.get(f"/proc/{pwndbg.aglib.proc.tid()}/personality")
            personality = int(data, 16)
            return (personality & 0x40000 == 0), "read status from process' personality"
        except Exception:
            print("Could not check ASLR: can't read process' personality")

    if not pwndbg.dbg.is_gdblib_available():
        return None, "Could not detect ASLR on LLDB"

    # Just go with whatever GDB says it did.
    #
    # This should usually be identical to the above, but we may not have
    # access to procfs.
    output = gdb.execute("show disable-randomization", to_string=True)
    return ("is off." in output), "show disable-randomization"


options = {"on": "off", "off": "on"}

parser = argparse.ArgumentParser(
    description="""
Check the current ASLR status, or turn it on/off.

Does not take effect until the program is restarted.
""",
)
parser.add_argument(
    "state",
    nargs="?",
    type=str,
    choices=options,
    help="Turn ASLR on or off (takes effect when target is started)",
)


@pwndbg.commands.Command(parser, category=CommandCategory.LINUX)
def aslr(state=None) -> None:
    if state:
        if pwndbg.dbg.is_gdblib_available():
            gdb.execute(
                f"set disable-randomization {options[state]}", from_tty=False, to_string=True
            )

            if pwndbg.aglib.proc.alive():
                print("Change will take effect when the process restarts")
        else:
            # TODO: lldb settings set target.disable-aslr false
            print(
                "Please use command 'settings set target.disable-aslr true/false', autocommand not supported yet"
            )

    aslr, method = check_aslr()

    if aslr is True:
        status = message.on("ON")
    elif aslr is False:
        status = message.off("OFF")
    else:
        status = message.off("???")

    print(f"ASLR is {status} ({method})")



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

from capstone6pwndbg import *  # noqa: F403
from pwnlib.constants import linux

import pwndbg.aglib
import pwndbg.aglib.memory
import pwndbg.aglib.remote
import pwndbg.aglib.typeinfo
import pwndbg.aglib.vmmap
import pwndbg.chain
import pwndbg.color.context as ctx_color
import pwndbg.color.memory as mem_color
import pwndbg.color.message as message
import pwndbg.color.syntax_highlight as H
import pwndbg.dintegration
import pwndbg.lib.config
import pwndbg.lib.disasm.helpers as bit_math
from pwndbg.aglib.disasm.instruction import FORWARD_JUMP_GROUP
from pwndbg.aglib.disasm.instruction import EnhancedOperand
from pwndbg.aglib.disasm.instruction import InstructionCondition
from pwndbg.aglib.disasm.instruction import PwndbgInstruction
from pwndbg.aglib.disasm.instruction import boolean_to_instruction_condition
from pwndbg.lib.arch import PWNDBG_SUPPORTED_ARCHITECTURES_TYPE
from pwndbg.lib.regs import PseudoEmulatedRegisterFile

# Emulator currently requires GDB, and we only use it here for type checking.
if TYPE_CHECKING:
    from pwndbg.emu.emulator import Emulator

pwndbg.config.add_param(
    "emulate",
    "on",
    "unicorn emulation of code from the current PC register",
    help_docstring="""\
Emulate can be:

1. off             - no emulation is performed
2. jumps-only      - emulation is done only to resolve branch instructions
3. on              - emulation is done to resolve registers/memory values etc.

Emulation can slow down Pwndbg. Disabling it may improve performance.
Emulation requires >1GB RAM being available on the system and ability to allocate RWX memory.
""",
    param_class=pwndbg.lib.config.PARAM_ENUM,
    enum_sequence=["on", "off", "jumps-only"],
)


# Even if this is disabled, branch instructions will still have targets printed
pwndbg.config.add_param(
    "disasm-annotations",
    True,
    "display annotations for instructions",
)

pwndbg.config.add_param(
    "emulate-annotations",
    True,
    "unicorn emulation for instruction annotations",
    help_docstring="Refers to register and memory value annotations.",
)

# If this is false, emulation is only used for the current instruction (if emulate-annotations is enabled)
pwndbg.config.add_param(
    "emulate-future-annotations",
    True,
    "unicorn emulation for future instruction's annotations",
)

# Effects future instructions, as past ones have already been cached and reflect the process state at the time
pwndbg.config.add_param("disasm-telescope-depth", 3, "depth of telescope for disasm annotations")

# In disasm view, long telescoped strings might cause lines wraps
pwndbg.config.add_param(
    "disasm-telescope-string-length",
    50,
    "the number of characters in strings to display in disasm annotations",
)

pwndbg.config.add_param(
    "disasm-inline-symbols",
    True,
    "replacing constant operands with their symbol in the disassembly",
)


def syntax_highlight(ins):
    return H.syntax_highlight(ins, filename=".asm")


DEBUG_ENHANCEMENT = False
# DEBUG_ENHANCEMENT = True

groups = {v: k for k, v in globals().items() if k.startswith("CS_GRP_")}
ops = {v: k for k, v in globals().items() if k.startswith("CS_OP_")}
access = {v: k for k, v in globals().items() if k.startswith("CS_AC_")}

for value1, name1 in dict(access).items():
    for value2, name2 in dict(access).items():
        # novermin
        access.setdefault(value1 | value2, f"{name1} | {name2}")

# These instruction types should not be emulated through, either
# because they cannot be emulated without interfering (syscall, etc.)
# or because they change privilege levels.
# There is an additional check for CS_GRP_CALL specially in the enhancement code, which we stop at
DO_NOT_EMULATE = {
    CS_GRP_INT,
    CS_GRP_INVALID,
    CS_GRP_IRET,
    # Note that we explicitly do not include the PRIVILEGE category, since
    # we may be in kernel code, and privileged instructions are just fine
    # in that case.
    # capstone.CS_GRP_PRIVILEGE,
}


def register_assign(left: str, right: str) -> str:
    return f"{left} => {right}"


def memory_assign(left: str, right: str) -> str:
    return f"{left} <= {right}"


def memory_or_register_assign(left: str, right: str, mem_assign: bool) -> str:
    """
    Used when we don't know until runtime whether a codepath will annotate a register or memory location.
    """
    return memory_assign(left, right) if mem_assign else register_assign(left, right)


# Enhances disassembly with memory values & symbols by adding member variables to an instruction
# The only public method that should be called is "enhance"
# The enhance function is passed an instance of the Unicorn emulator
#  and will .single_step() it to determine operand values before and after executing the instruction
class DisassemblyAssistant:
    architecture: PWNDBG_SUPPORTED_ARCHITECTURES_TYPE

    manual_register_values: PseudoEmulatedRegisterFile

    supports_manual_emulation = False
    """This feature relies on the Capstone .regs_access() features that not all architectures have reliable support for"""

    def __init__(self, architecture: PWNDBG_SUPPORTED_ARCHITECTURES_TYPE) -> None:
        self.architecture = architecture
        self.manual_register_values = PseudoEmulatedRegisterFile(
            pwndbg.aglib.regs.current, pwndbg.aglib.arch.ptrsize
        )

        self.op_handlers: dict[
            int, Callable[[PwndbgInstruction, EnhancedOperand, Emulator], int | None]
        ] = {
            CS_OP_IMM: self._parse_immediate,  # Return immediate value
            CS_OP_REG: self._parse_register,  # Return value of register
            # Handler for memory references (as dictated by Capstone), such as first operand of "mov qword ptr [rbx + rcx*4], rax"
            CS_OP_MEM: self._parse_memory,  # Return parsed address, do not dereference
        }

        # Return a string corresponding to operand. Used to reduce code duplication while printing
        # REG type wil return register name, "RAX"
        self.op_names: dict[int, Callable[[PwndbgInstruction, EnhancedOperand], str | None]] = {
            CS_OP_IMM: self._immediate_string,
            CS_OP_REG: self._register_string,
            CS_OP_MEM: self._memory_string,
        }

    def enhance(self, instruction: PwndbgInstruction, emu: Emulator = None) -> None:
        """
        Enhance the instruction - resolving branch targets, conditionals, and adding annotations

        This is the only public method that should be called on this object externally.
        """
        # It is assumed that the emulator's pc is at the instruction's address

        # There are 3 degrees of emulation:
        # 1. No emulation at all. In this case, the `emu` parameter should be None
        # 2. Only emulate jumps - the only interaction with the emulator in this case is stepping it and reading the PC
        # 3. Full emulation - read registers and memory from the emulator as well as determining jumps

        if DEBUG_ENHANCEMENT:
            print(
                f"Start enhancing instruction at {hex(instruction.address)} - {instruction.mnemonic} {instruction.op_str}"
            )

        # Get another reference to the emulator for the purposes of jumps
        jump_emu = emu

        if pwndbg.config.emulate != "on":
            emu = None

        # For both cases below, set emu to None so we don't use it for annotation
        if emu and not bool(pwndbg.config.emulate_annotations):
            emu = None

        # Disable emulation for future annotations based on setting
        if (
            emu
            and pwndbg.aglib.regs.pc != instruction.address
            and not bool(pwndbg.config.emulate_future_annotations)
        ):
            emu = None

        # Ensure emulator's program counter is at the correct location.
        # This occurs very rarely - observed sometimes when the remote is stalling, ctrl-c, and for some reason emulator returns PC=0.
        if emu:
            if emu.pc() != instruction.address:
                if DEBUG_ENHANCEMENT:
                    print(
                        f"Program counter and emu.pc do not line up: {hex(pwndbg.aglib.regs.pc)=} {hex(emu.pc())=}"
                    )
                emu = jump_emu = None

        self._prepare(instruction, emu)

        # Don't disable emulation yet, as we can use it to read the syscall register
        self._enhance_syscall(instruction, emu)

        # Disable emulation for instructions we don't want to emulate (CALL, INT, ...)
        if emu and set(instruction.groups) & DO_NOT_EMULATE:
            emu.valid = False
            emu = jump_emu = None

            if DEBUG_ENHANCEMENT:
                print("Turned off emulation - not emulating certain type of instruction")

        # This function will .single_step the emulation
        if not self._enhance_operands(instruction, emu, jump_emu):
            if jump_emu is not None and DEBUG_ENHANCEMENT:
                print(f"Emulation failed at {instruction.address=:#x}")
            emu = None
            jump_emu = None

        if jump_emu is not None:
            # We successfully used emulation for this instruction
            instruction.emulated = True

        # Set the .condition field
        self._enhance_conditional(instruction, emu)

        # Set the .target and .next fields
        self._enhance_next(instruction, emu, jump_emu)

        if (
            bool(pwndbg.config.disasm_annotations)
            and instruction.condition != InstructionCondition.FALSE
        ):
            self._set_annotation_string(instruction, emu)

        # Disable emulation after CALL instructions. We do it after enhancement, as we can use emulation
        # to determine the call's target address.
        if jump_emu and instruction.call_like:
            jump_emu.valid = False
            jump_emu = None
            emu = None

            if DEBUG_ENHANCEMENT:
                print("Turned off emulation for call")

        # Manually propagate register values so when enhancing the next instruction, we can read from these registers
        if self.supports_manual_emulation:
            if (
                instruction.call_like
                or (set(instruction.groups) & DO_NOT_EMULATE)
                or (instruction.jump_like and not instruction.jump_result_is_known)
            ):
                # Syscalls and functions (which we step over) can clobber registers
                # Also, if we encounter a control flow instruction where the result is unknown,
                # we need to reset the registers because otherwise it may show annotations for instructions never actually taken.
                self.manual_register_values.invalidate_all_registers()
            else:
                _, regs_written = instruction.cs_insn.regs_access()

                for reg_id in regs_written:
                    reg_name: str = instruction.cs_insn.reg_name(reg_id)

                    # If we determined that this instruction wrote some value to this register, propagate it.
                    # Otherwise, invalidate the value since we cannot reason about it.
                    if reg_id in instruction.register_writes:
                        self.manual_register_values.write_register(
                            reg_name, instruction.register_writes[reg_id]
                        )
                    else:
                        self.manual_register_values.invalidate_register(reg_name)

        if DEBUG_ENHANCEMENT:
            print(self.dump(instruction))
            print("Done enhancing")

    # This is run before enhancement - often used to handle edge case behavior
    def _prepare(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        return None

    # Subclasses for specific architecture should override this
    def _set_annotation_string(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        """
        The goal of this function is to set the `annotation` field of the instruction,
        which is the string to be printed in a disasm view.
        """
        return None

    def _enhance_operands(
        self, instruction: PwndbgInstruction, emu: Emulator, jump_emu: Emulator
    ) -> bool:
        """
        Enhances the operands by determining values and symbols

        When emulation is enabled, this will `single_step` the emulation to determine the value of registers
        before and after the instrution has executed.

        For each operand explicitly written to or read from (instruction.operands), sets the following fields:

            operand.before_value
                Integer value of the operand before instruction executes.
                None if cannot be resolved/reasoned about.

            operand.after_value
                Integer value of the operand after instruction executes.
                Only set when emulation is enabled. Otherwise None.
                This is relevent if we read and write to the same registers within an instruction

            operand.symbol:
                Resolved symbol name for this operand, if .before_value is set, else None.

            operand.str:
                String representing the operand

        Return False if emulation fails (so we don't use it in additional enhancement steps)
        """

        # Apply syntax highlighting to the assembly
        if pwndbg.config.syntax_highlight:
            instruction.asm_string = syntax_highlight(instruction.asm_string)

        stack_vars = pwndbg.dintegration.manager.get_stack_var_dict_all()

        # Populate the "operands" list of the instruction
        # Set before_value, symbol, and str
        for op in instruction.operands:
            # Retrieve the value, either an immediate, from a register, or from memory
            op.before_value = self.op_handlers.get(op.type, lambda *a: None)(instruction, op, emu)
            if op.before_value is not None:
                # Don't mask immediates - some computations depend on their signed values
                if op.type is not CS_OP_IMM:
                    op.before_value &= pwndbg.aglib.arch.ptrmask

                if op.before_value >= 0:
                    op.symbol = mem_color.attempt_colorized_symbol(op.before_value, stack_vars)

                op.before_value_resolved = self._resolve_used_value(
                    op.before_value, instruction, op, emu
                )

                if (
                    op.symbol
                    and op.type in (CS_OP_IMM, CS_OP_MEM)
                    and pwndbg.config.disasm_inline_symbols
                ):
                    # Make an inline replacement, so `jmp 0x400122` becomes `jmp function_name`
                    instruction.asm_string = instruction.asm_string.replace(
                        hex(op.before_value), op.symbol
                    )

        # Execute the instruction
        if jump_emu and None in jump_emu.single_step(instruction=instruction):
            # This branch is taken if stepping the emulator failed
            jump_emu = None
            emu = None

        # Set after_value after single stepping the emulator
        if emu is not None:
            # after_value
            for op in instruction.operands:
                # Retrieve the value, either an immediate, from a register, or from memory
                op.after_value = self.op_handlers.get(op.type, lambda *a: None)(
                    instruction, op, emu
                )

                op.after_value_resolved = self._resolve_used_value(
                    op.after_value, instruction, op, emu
                )

                if op.after_value is not None:
                    op.after_value &= pwndbg.aglib.arch.ptrmask

        # Set .str value of operands, after emulation has been completed
        for op in instruction.operands:
            op.str = self.op_names.get(op.type, lambda *a: None)(instruction, op)

        return jump_emu is not None

    def can_reason_about_process_state(self, instruction: PwndbgInstruction) -> bool:
        """
        Determine if the program counter of the process equals the address of the instruction being enhanced.
        If so, it means we can safely reason and read from registers and memory to enhance values that
        we can add to the annotation string. This becomes relevent when NOT emulating, and is meant to
        allow more details when the PC is at the instruction being enhanced
        """
        return instruction.address == pwndbg.aglib.regs.pc

    # Delegates to "read_register", which takes Capstone ID for register.
    def _parse_register(
        self, instruction: PwndbgInstruction, operand: EnhancedOperand, emu: Emulator
    ) -> int | None:
        reg = operand.reg
        return self._read_register(instruction, reg, emu)

    # Determine memory address of operand (Ex: in x86, mov rax, [rip + 0xd55], would return $rip_after_instruction+0xd55)
    # Subclasses override for specific architectures
    def _parse_memory(
        self, instruction: PwndbgInstruction, operand: EnhancedOperand, emu: Emulator
    ) -> int | None:
        return None

    def _parse_immediate(
        self, instruction: PwndbgInstruction, operand: EnhancedOperand, emu: Emulator
    ):
        return operand.imm

    def _read_register(
        self, instruction: PwndbgInstruction, operand_id: int, emu: Emulator
    ) -> int | None:
        """
        Read value in register. Return None if cannot reason about the value in the register.
        Different architectures use registers in different patterns, so it is best to
        override this to get to best behavior for a given architecture. See x86.py as example.

        operand_id is the ID internal to Capstone
        """
        regname: str = instruction.cs_insn.reg_name(operand_id)
        return self._read_register_name(instruction, regname, emu)

    # Read register by its name
    def _read_register_name(
        self, instruction: PwndbgInstruction, regname: str, emu: Emulator
    ) -> int | None:
        if emu:
            # Will read the value of register from the emulator
            # Be concious about calling this before/after stepping the emulator
            value = emu.read_register(regname)
            if DEBUG_ENHANCEMENT:
                print(f"Register in emulation returned {regname}={hex(value)}")
            return value
        if self.can_reason_about_process_state(instruction):
            # When instruction address == pc, we can reason about all registers.
            # The values will just reflect values prior to executing the instruction, instead of after,
            # which is relevent if we are writing to this register.
            # However, the information can still be useful for display purposes.
            if DEBUG_ENHANCEMENT:
                print(f"Read value from process register: {pwndbg.aglib.regs.read_reg(regname)}")
            return pwndbg.aglib.regs.read_reg(regname)
        if (reg_value := self.manual_register_values.read_register(regname)) is not None:
            # If we manually tracked the value of this register while disassembling, we can read from it.
            return reg_value
        return None

    # Read memory of given size, taking into account emulation and being able to reason about the memory location
    def _read_memory(
        self,
        address: int,
        size: int,
        instruction: PwndbgInstruction,
        emu: Emulator,
    ) -> int | None:
        address_list = self._telescope(address, 1, instruction, emu, read_size=size)

        if len(address_list) >= 2:
            return address_list[1]

        return None

    # Pass in a operand and it's value, and determine the actual value used during an instruction
    # Helpful for cases like  `cmp    byte ptr [rip + 0x166669], 0`, where first operand could be
    # a register or a memory value to dereference, and we want the actual value used.
    # Override this to implement memory lookups in given architecture (if it's relevent)
    # Different architecture read memory differently:
    # - Only a couple Capstone architectures support the memory .size field, which determines read width.
    # - In others, read/write width is implied.
    def _resolve_used_value(
        self,
        value: int | None,
        instruction: PwndbgInstruction,
        operand: EnhancedOperand,
        emu: Emulator,
    ) -> int | None:
        if value is None:
            return None

        if operand.type == CS_OP_REG or operand.type == CS_OP_IMM:
            return value
        if operand.type == CS_OP_MEM:
            # Assume that we are reading ptrsize - subclasses should override this function
            # to provide a more specific value if needed
            self._read_memory(value, pwndbg.aglib.arch.ptrsize, instruction, emu)

        return None

    def _telescope(
        self,
        address: int,
        limit: int,
        instruction: PwndbgInstruction,
        emu: Emulator,
        read_size: int = None,
    ) -> list[int]:
        """
        Dereference an address recursively - takes into account emulation.

        It will only dereference as it is safe to do so, meaning the last value in the returned list may be a pointer

        The list that the function returns is guaranteed have len >= 1
        """

        can_read_process_state = self.can_reason_about_process_state(instruction)

        if emu:
            return emu.telescope(address, limit, read_size=read_size)
        if can_read_process_state:
            # Can reason about memory in this case.

            if read_size is not None and read_size < pwndbg.aglib.arch.ptrsize:
                result = [address]

                size_type = pwndbg.aglib.typeinfo.get_type(read_size)
                try:
                    read_value = int(
                        pwndbg.aglib.memory.get_typed_pointer_value(size_type, address)
                    )
                    result.append(read_value)
                except pwndbg.dbg_mod.Error:
                    pass

                return result

            return pwndbg.chain.get(address, limit=limit)
        # If the target address is in a non-writeable map, we can pretty safely telescope
        # This is best-effort to give a better experience

        address_list = [address]

        if read_size is not None and read_size < pwndbg.aglib.arch.ptrsize:
            size_type = pwndbg.aglib.typeinfo.get_type(read_size)
        else:
            size_type = pwndbg.aglib.typeinfo.ppvoid

        for _ in range(limit):
            if address_list.count(address) >= 2:
                break

            page = pwndbg.aglib.vmmap.find(address)
            if page and not page.write:
                try:
                    address = int(pwndbg.aglib.memory.get_typed_pointer_value(size_type, address))
                    address &= pwndbg.aglib.arch.ptrmask
                    address_list.append(address)
                except pwndbg.dbg_mod.Error:
                    break
            else:
                break

        return address_list

        # We cannot telescope, but we can still return the address.
        # Just without any further information
        return [address]

    # Dispatch to the appropriate format handler. Pass the list returned by `telescope()` to this function
    def _telescope_format_list(self, addresses: list[int], limit: int, emu: Emulator) -> str:
        # It is assumed proper checks have been made BEFORE calling this function so that pwndbg.chain.format
        #  will return values accurate to the program state at the time of instruction executing.

        enhance_string_len = int(pwndbg.config.disasm_telescope_string_length)

        if emu:
            return emu.format_telescope_list(
                addresses, limit, enhance_string_len=enhance_string_len
            )
        # We can format, but in some cases we may not be able to reason about memory, so don't allow
        # it to dereference to last value in memory (we can't determine what value it is)
        return pwndbg.chain.format(
            addresses,
            limit=limit,
            enhance_string_len=enhance_string_len,
        )

    @staticmethod
    def _syscall_name(number: int, arch: str) -> str | None:
        """
        Given a syscall number and architecture, returns the name of the syscall.
        E.g. execve == 59 on x86-64
        """
        arch_module = {
            "arm": linux.arm,
            "armcm": linux.arm,
            "i386": linux.i386,
            "mips": linux.mips,
            "x86-64": linux.amd64,
            "aarch64": linux.aarch64,
            "rv32": linux.riscv64,
            "rv64": linux.riscv64,
        }.get(arch)

        if arch_module is None:
            return None

        prefix = "__NR_"

        for k, v in arch_module.__dict__.items():
            if v != number:
                continue

            if not k.startswith(prefix):
                continue

            return k[len(prefix) :].lower()

        return None

    def _enhance_syscall(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        if CS_GRP_INT not in instruction.groups:
            return None

        syscall_arch, syscall_register = self._get_syscall_arch_info(instruction)

        if syscall_arch is None:
            return None

        instruction.syscall = self._read_register_name(instruction, syscall_register, emu)
        if instruction.syscall is not None:
            instruction.syscall_name = (
                DisassemblyAssistant._syscall_name(instruction.syscall, syscall_arch)
                or f"<unk_{instruction.syscall}>"
            )

    def _get_syscall_arch_info(self, instruction) -> tuple[str, str]:
        """
        Return tuple of (name of syscall architecture, syscall register name)

        Elements of the tuple will be None to indicate it's not a syscall
        """
        if pwndbg.aglib.arch.syscall_abi is None:
            return (None, None)
        return (pwndbg.aglib.arch.name, pwndbg.aglib.arch.syscall_abi.syscall_register)

    def _enhance_conditional(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        """
        Sets the `condition` of the instruction

        If the instruction is always executed unconditionally, or we cannot reason about the instruction,
        the value of the field is `InstructionCondition.UNDETERMINED`.

        If the instruction is executed conditionally, and we can be absolutely
        sure that it will be executed, the value of the field is `InstructionCondition.TRUE`.

        In all other cases, it is set to `InstructionCondition.FALSE`.
        """

        instruction.condition = self._condition(instruction, emu)

    # Subclasses should override
    def _condition(self, instruction: PwndbgInstruction, emu: Emulator) -> InstructionCondition:
        return InstructionCondition.UNDETERMINED

    def _enhance_next(
        self, instruction: PwndbgInstruction, emu: Emulator, jump_emu: Emulator
    ) -> None:
        """
        Set the `next` and `target` field of the instruction.

        By default, it is set to the address of the next linear
        instruction.

        `next` is the address that the PC would be upon using the GDB `nexti` command,
        `target` is the jump target whether or not the jump is taken, like `stepi` and assuming the jump is taken.

        If the instruction is a non-"call" branch and either:
        - Is unconditional, or is conditional and is known to be taken, a
        - Is conditional, but is known to be taken

        And the target can be resolved, it is set to the address
        of the jump target.

        """
        next_addr: int | None = None

        # The order for the following statements in determining the next executed instruction is important
        #
        # Firstly, we check the condition field - this field is manually set by our enhancement code
        # There are cases where the Unicorn emulator is incorrect - for example, delay slots in MIPS causing jumps to not resolve correctly
        # due to the way we single-step the emulator. We want our own manual checks to override the emulator

        if not instruction.call_like and (
            instruction.condition == InstructionCondition.TRUE or instruction.is_unconditional_jump
        ):
            # Don't allow call instructions - we want the actual "nexti" address
            # If condition is true, then this might be a conditional jump
            # There are some other instructions that run conditionally though - resolve_target returns None in those cases
            # Or, if this is a unconditional jump, we will try to resolve target
            next_addr = self._resolve_target(instruction, emu)

        # Secondly, attempt to use emulation if we could not resolve the target above, or don't have custom condition handler for the architecture yet
        if next_addr is None and jump_emu:
            # Use emulator to determine the next address:
            # 1. Only use it to determine non-call's (`nexti` should step over calls)
            # 2. Make sure we haven't manually set .condition to False (which should override the emulators prediction)
            if not instruction.call_like and instruction.condition != InstructionCondition.FALSE:
                next_addr = jump_emu.pc()

        # Handle edge case - if the target happens to be the next address in memory and it's a jump, we need this variable
        # so the disasm output is accurate.
        if next_addr is not None and instruction.is_unconditional_jump:
            instruction.force_unconditional_jump_target = True

        # All else fails, take the next instruction in memory
        if next_addr is None:
            next_addr = instruction.address + instruction.size

        # Determine the target of this address.
        # This is the address that the instruction could potentially change the program counter to, meaning that `stepi` would go to the target
        instruction.target = self._resolve_target(instruction, emu)

        instruction.next = next_addr & pwndbg.aglib.arch.ptrmask

        if instruction.target is None:
            instruction.target = instruction.next

        if instruction.has_jump_target and instruction.target >= 0:
            # Only bother doing the symbol lookup if this is a jump
            instruction.target_string = mem_color.get_address_or_symbol(
                instruction.target, pwndbg.dintegration.manager.get_stack_var_dict_all()
            )

        # Now that we have determined the target, if it was a conditional branch,
        # go back and correct the instruction condition to reflect the branch decision of the emulator
        # in case we didn't manually determine the condition.
        if (
            jump_emu
            and instruction.condition == InstructionCondition.UNDETERMINED
            and instruction.is_conditional_jump
        ):
            # At this point we know the emulator was used to determine
            # the conditional branch
            instruction.condition = boolean_to_instruction_condition(
                instruction.is_conditional_jump_taken
            )

        if (
            instruction.operands
            and instruction.operands[0].before_value
            and instruction.operands[0].type == CS_OP_IMM
        ):
            instruction.target_const = True

    # This is the default implementation.
    # Subclasses should override this for more accurate behavior/to catch more cases. See x86.py as example
    def _resolve_target(self, instruction: PwndbgInstruction, emu: Emulator | None):
        """
        Architecture-specific hook point for _enhance_next.

        Returns the program counter target of this instruction.
        Even in the case of conditional jumps, the potential target should be resolved.
        """

        # The FORWARD_JUMP_GROUP here is very specific
        # We only want this resolver to work for instructions that Capstone
        # explicitely labels as jump instructions. If we determine that another type of instruction
        # can have a target, we resolve it manually, as this manual resolver would return improper values,
        # as it is built on the assumptions of branch instructions across many architectures.
        if not bool(instruction.groups & FORWARD_JUMP_GROUP):
            return None

        addr = None

        # At this point, all operands have been resolved.
        # Assume only single-operand jumps.
        if len(instruction.operands) == 1:
            op = instruction.operands[0]
            addr = self._resolve_used_value(op.before_value, instruction, op, emu)
            if addr:
                addr &= pwndbg.aglib.arch.ptrmask
        else:
            # Some architectures have jumps with multiple operands. In this case, this default implementation
            # does a simple naive check. Iterate all operands, pick the first one resolves to a symbol or lands in executable memory
            # and use that as the target

            # Reversed order, just because through observation the immediates and labels are often farther right
            for op in reversed(instruction.operands):
                resolved_addr = self._resolve_used_value(op.before_value, instruction, op, emu)
                if resolved_addr:
                    resolved_addr &= pwndbg.aglib.arch.ptrmask
                    if op.symbol:
                        addr = resolved_addr
                    else:
                        page = pwndbg.aglib.vmmap.find(resolved_addr)
                        # When debugging a remote QEMU target, the page permissions are not accurate.
                        # In this case, if the candidate address is mapped at all, just go with it.
                        if page and (page.execute or pwndbg.aglib.remote.is_remote()):
                            addr = resolved_addr

                if addr is not None:
                    instruction.target_const = op.type == CS_OP_IMM
                    break

        if addr is None:
            return None

        return int(addr)

    def dump(self, instruction: PwndbgInstruction):
        """
        Debug-only method.
        """
        return repr(instruction)

    # String functions assume the .before_value and .after_value have been set
    def _immediate_string(self, instruction, operand) -> str:
        return pwndbg.lib.pretty_print.int_to_string(operand.before_value)

    def _register_string(self, instruction: PwndbgInstruction, operand: EnhancedOperand):
        """
        Return colorized register string
        """
        reg = operand.reg
        name = ctx_color.register(instruction.cs_insn.reg_name(reg).upper())

        # If using emulation and we determined the value didn't change, don't colorize
        if (
            operand.before_value is not None
            and operand.after_value is not None
            and operand.before_value == operand.after_value
        ):
            return name
        return ctx_color.register_changed(name)

    def _memory_string(self, instruction: PwndbgInstruction, operand: EnhancedOperand):
        """
        Example: return "[_IO_2_1_stdin_+16]", where the address/symbol is colorized
        """
        if operand.before_value is not None:
            return f"[{mem_color.get_address_or_symbol(operand.before_value, pwndbg.dintegration.manager.get_stack_var_dict_all())}]"
        return None

    def _common_generic_register_destination(
        self, instruction: PwndbgInstruction, emu: Emulator
    ) -> None:
        """
        This function can be used to annotate instructions that have a register destination.
        In the vast majority of instructions in most architectures, the destination register is the first operand.

        Using emulation, it will determine the value placed into the register, and create an annotation string based on the result.
        """

        left = instruction.operands[0]

        # Emulating determined the value that was set in the destination register
        if left.after_value is not None:
            TELESCOPE_DEPTH = max(0, int(pwndbg.config.disasm_telescope_depth))

            # Telescope the address
            telescope_addresses = self._telescope(
                left.after_value,
                TELESCOPE_DEPTH + 1,
                instruction,
                emu,
                read_size=pwndbg.aglib.arch.ptrsize,
            )

            if not telescope_addresses:
                return

            instruction.annotation = register_assign(
                left.str, self._telescope_format_list(telescope_addresses, TELESCOPE_DEPTH, emu)
            )

    def _common_cmp_annotator_builder(
        self, flags_register_name: str, char_to_separate_operands: str = "-"
    ) -> Callable[[PwndbgInstruction, Emulator], None]:
        """
        Many architectures implement near-identical `CMP`-like instructions.

        It takes two values, either subtracts, adds, or does some bit operation
        with them to set values in the flag register.

        To reduce code duplication, subclasses can use this function to create an annotator for CMP-like instructions.
        """
        FLAG_REG_NAME_DISPLAY = flags_register_name.upper()

        def handler(instruction: PwndbgInstruction, emu: Emulator):
            # If there are just two operands, we can assume we are comparing them directly, and can display the values.
            # Some architectures have variants with more operands.
            if len(instruction.operands) == 2:
                left, right = instruction.operands

                if (l_value := left.before_value_resolved) is not None and (
                    r_value := right.before_value_resolved
                ) is not None:
                    print_left, print_right = pwndbg.lib.pretty_print.int_pair_to_string(
                        l_value, r_value
                    )
                    # Ex: "0x7f - 0x12" or "0xdffffdea + 0x8"
                    instruction.annotation = (
                        f"{print_left} {char_to_separate_operands} {print_right}"
                    )

            # Using emulation, we can determine the resulting value put into the flag register
            if emu:
                eflags_bits = pwndbg.aglib.regs.flags[flags_register_name]
                emu_eflags = emu.read_register(flags_register_name)
                eflags_formatted = ctx_color.format_flags(emu_eflags, eflags_bits)

                display_result = register_assign(FLAG_REG_NAME_DISPLAY, eflags_formatted)

                if instruction.annotation is None:
                    # First part of this function usually sets .annotation to a string. But if the instruction
                    # has more than two operands, then we don't have a way of showing them, so this avoids the "+="" below
                    instruction.annotation = display_result
                else:
                    instruction.annotation += " " * 5 + display_result

        return handler

    def _common_load_annotator(
        self,
        instruction: PwndbgInstruction,
        emu: Emulator,
        address: int | None,
        read_size: int,
        signed: bool,
        target_size: int,
        dest_str: str,
        source_str: str,
    ) -> None:
        """
        This function annotates load instructions - moving data from memory into a register.

        These instructions read `read_size` bytes from memory into a register.

        `signed`: whether or not we are loading a signed value from memory
        `target_size`: the size of the register in bytes - relevent for sign-extension
        `dest_str`: a string representing the destination register ('rax')
        `source_str`: a string representing the source address ('[0x7fffffffe138]')
        """

        if address is None:
            return

        # There are many cases we need to consider when we are loading a value from memory
        # Were we able to reason about the memory address, and dereference it?
        # Does the resolved memory address actual point into memory?
        # If the target register size is larger than the read size, then do we need sign-extension?

        # If the address is not mapped, we segfaulted
        if not pwndbg.aglib.memory.peek(address):
            instruction.annotation = message.error(
                f"<Cannot dereference [{mem_color.get(address)}]>"
            )
        else:
            # In this branch, it is assumed that the address IS in a mapped page
            TELESCOPE_DEPTH = max(1, int(pwndbg.config.disasm_telescope_depth))

            telescope_addresses = self._telescope(
                address,
                TELESCOPE_DEPTH,
                instruction,
                emu,
                read_size=read_size,
            )

            if len(telescope_addresses) == 1:
                # If telescope returned only 1 address (and we already know the address is in a mapped page)
                # it means we couldn't reason about the dereferenced memory.
                # In this case, simply display the address

                # As an example, this path is taken for the following case:
                # mov rdi, qword ptr [rip + 0x17d40] where the resolved memory address is in writeable memory,
                # and we are not emulating. This means we cannot savely dereference if PC is not at the current instruction address,
                # because the the memory address could have been written to by the time the instruction executes
                telescope_print = None
            else:
                if signed and read_size != target_size and len(telescope_addresses) == 2:
                    # We sign extend the value, then convert it back to the unsigned bit representation
                    final_value = bit_math.to_signed(telescope_addresses[1], read_size * 8) & (
                        (1 << (target_size * 8)) - 1
                    )
                    # If it's a signed read that required extension, it will just be a number with no special symbol/color needed
                    telescope_print = hex(final_value)
                else:
                    # Start showing at dereferenced address, hence the [1:]
                    telescope_print = f"{self._telescope_format_list(telescope_addresses[1:], TELESCOPE_DEPTH, emu)}"

            instruction.annotation = f"{dest_str}, {source_str}"

            if telescope_print is not None:
                instruction.annotation = register_assign(instruction.annotation, telescope_print)

    def _common_store_annotator(
        self,
        instruction: PwndbgInstruction,
        emu: Emulator,
        address: int | None,
        value: int | None,
        write_size: int | None,
        address_str: str,
    ) -> None:
        """
        This function annotates store functions - moving data from a register to memory.

        The `value` is truncated to match the `write_size`, if `write_size` is not None.

        The annotation will indicate if the instruction will segfault.

        `write_size`: number of bytes of `value` that will be written
        """

        if address is None:
            return

        if not pwndbg.aglib.memory.peek(address):
            instruction.annotation = message.error(
                f"<Cannot dereference [{mem_color.get(address)}]>"
            )
        elif value is not None:
            # To make this annotation work with emulation disabled,
            # we telescope the value that is going to be placed in the memory operand
            TELESCOPE_DEPTH = max(0, int(pwndbg.config.disasm_telescope_depth))

            if write_size is not None:
                value &= (1 << (write_size * 8)) - 1

            telescope_addresses = self._telescope(
                value,
                TELESCOPE_DEPTH,
                instruction,
                emu,
            )

            instruction.annotation = memory_assign(
                address_str, self._telescope_format_list(telescope_addresses, TELESCOPE_DEPTH, emu)
            )

    def _common_move_annotator(self, instruction: PwndbgInstruction, emu: Emulator):
        """
        This function handles annotating `MOV` type instructions - where the value of one register is placed into another.
        """
        if len(instruction.operands) == 2:
            left, right = instruction.operands
            # If we already used emulation, use the result, otherwise take the source operand before_value
            result = left.after_value or right.before_value
            if result is not None and result >= 0:
                # We have determined the value written to this register - propagate this to future instructions.
                instruction.register_writes[left.reg] = result

                TELESCOPE_DEPTH = max(0, int(pwndbg.config.disasm_telescope_depth))

                telescope_addresses = self._telescope(
                    result,
                    TELESCOPE_DEPTH + 1,
                    instruction,
                    emu,
                )
                if not telescope_addresses:
                    return

                instruction.annotation = register_assign(
                    left.str, self._telescope_format_list(telescope_addresses, TELESCOPE_DEPTH, emu)
                )

    def _common_binary_op_annotator(
        self,
        instruction: PwndbgInstruction,
        emu: Emulator,
        target_operand: EnhancedOperand,
        op_one: int | None,
        op_two: int | None,
        char_to_separate_operands: str,
        memory_assignment=False,
    ) -> None:
        # Ex: "0x198723 + 0x2b8"
        math_string = None

        if op_one is not None and op_two is not None:
            print_left, print_right = pwndbg.lib.pretty_print.int_pair_to_string(op_one, op_two)

            math_string = f"{print_left} {char_to_separate_operands} {print_right}"

        # Using emulation, we can determine the resulting value
        if target_operand.after_value_resolved is not None:
            instruction.annotation = memory_or_register_assign(
                target_operand.str,
                mem_color.get_address_and_symbol(
                    target_operand.after_value_resolved,
                    pwndbg.dintegration.manager.get_stack_var_dict_all(),
                ),
                memory_assignment,
            )
            if math_string:
                instruction.annotation += f" ({math_string})"
        elif math_string:
            instruction.annotation = memory_or_register_assign(
                target_operand.str, math_string, memory_assignment
            )


def basic_enhance(ins: PwndbgInstruction) -> None:
    # Apply syntax highlighting and inline symbol replacement
    # Used in cases were we don't want to do the full enhancement process
    # for performance reasons.
    if pwndbg.config.syntax_highlight:
        ins.asm_string = syntax_highlight(ins.asm_string)

    if pwndbg.config.disasm_inline_symbols:
        stack_vars = pwndbg.dintegration.manager.get_stack_var_dict_all()
        # Make inline replacements, so `jmp 0x400122` becomes `jmp function_name`
        for op in ins.operands:
            if op.type is CS_OP_IMM:
                op.before_value = op.imm

                if op.before_value >= 0:
                    op.symbol = mem_color.attempt_colorized_symbol(op.before_value, stack_vars)

                if op.symbol:
                    ins.asm_string = ins.asm_string.replace(hex(op.before_value), op.symbol)



"""
Getting Thread Local Storage (TLS) information.
"""

from __future__ import annotations

import pwndbg.aglib
import pwndbg.aglib.symbol
import pwndbg.aglib.typeinfo
import pwndbg.dbg_mod


def __call_pthread_self() -> int:
    """
    Retrieve the address of the `struct pthread_t` for the current thread by
    calling pthread_self(). This address can be used to locate the base address
    of the Thread Local Storage (TLS).
    """
    if pwndbg.aglib.symbol.lookup_symbol_addr("pthread_self") is None:
        return 0
    try:
        return int(
            pwndbg.dbg.selected_frame().evaluate_expression(
                "(void *)pthread_self()", lock_scheduler=True
            )
        )
    except pwndbg.dbg_mod.Error:
        return 0


def find_address_with_pthread_self() -> int:
    """
    Get the base address of the Thread Local Storage (TLS) for the current thread using
    the pthread_self() function. The returned address points to the `struct tcbhead_t`,
    which serves as the header for TLS and thread-specific metadata.
    """
    if pwndbg.aglib.arch.name not in ("x86-64", "i386", "arm", "aarch64", "loongarch64"):
        return 0

    result = __call_pthread_self()
    if result <= 0:
        # pthread_self() is not valid
        return 0

    # pthread_self() is defined as: https://elixir.bootlin.com/glibc/glibc-2.37/source/nptl/pthread_self.c#L22
    # THREAD_SELF is defined as:
    # i386: https://elixir.bootlin.com/glibc/glibc-2.37/source/sysdeps/i386/nptl/tls.h#L234
    # x86-64: https://elixir.bootlin.com/glibc/glibc-2.37/source/sysdeps/x86_64/nptl/tls.h#L181
    # arm: https://elixir.bootlin.com/glibc/latest/source/sysdeps/arm/nptl/tls.h#L76
    # loongarch64: https://github.com/bminor/glibc/blob/1c9ac027a5deb6c3e026be0e88d38959529e6102/sysdeps/loongarch/nptl/tls.h#L64
    # For i386 and x86-64, the return value of the pthread_self() is the address of TLS, because the value is self reference of the TLS: https://elixir.bootlin.com/glibc/glibc-2.37/source/nptl/pthread_create.c#L671
    # But for arm, the implementation of THREAD_SELF is different, we need to add sizeof(struct pthread) to the result to get the address of TLS.

    if pwndbg.aglib.arch.name in ("arm", "aarch64"):
        pthread_type = pwndbg.aglib.typeinfo.load("struct pthread")
        if pthread_type is None:
            # Type 'pthread' not found
            return 0
        result += pthread_type.sizeof
    elif pwndbg.aglib.arch.name == "loongarch64":
        pthread_type = pwndbg.aglib.typeinfo.load("struct pthread")
        if pthread_type is None:
            # Type 'pthread' not found
            return 0
        result += pthread_type.sizeof + pthread_type.alignof

    return result


def find_address_with_register() -> int:
    """
    Get the base address of the Thread Local Storage (TLS) for the current thread using
    a CPU register. The returned address points to the `struct tcbhead_t`, which is the
    entry point for TLS and thread-specific metadata.
    """
    if pwndbg.aglib.arch.name == "x86-64":
        return int(pwndbg.aglib.regs.fsbase)
    if pwndbg.aglib.arch.name == "i386":
        return int(pwndbg.aglib.regs.gsbase)
    if pwndbg.aglib.arch.name == "aarch64":
        # FIXME: cleanup/remove `TPIDR_EL0` register, it was renamed to `tpidr` since GDB13+
        return int(
            pwndbg.aglib.regs.read_reg("tpidr") or pwndbg.aglib.regs.read_reg("TPIDR_EL0") or 0
        )
    if pwndbg.aglib.arch.name == "arm":
        # TODO: linux ptrace for 64bit kernel?
        # In FreeBSD tls is under `tpidruro` register.
        # In Linux, the `tpidruro` register isn't available via ptrace in the 32-bit
        # kernel but it is available for an aarch32 program running under an arm64
        # kernel via the ptrace compat interface.
        return int(pwndbg.aglib.regs.read_reg("tpidruro") or 0)
    if pwndbg.aglib.arch.name == "loongarch64":
        return int(pwndbg.aglib.regs.read_reg("tp") or 0)
    return 0



from __future__ import annotations

import argparse
import shutil
import sys
import urllib.error
import urllib.request
from pathlib import Path

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.symbol
import pwndbg.color as color
import pwndbg.color.memory as color_mem
import pwndbg.color.message as message
import pwndbg.commands
import pwndbg.dbg_mod
import pwndbg.dintegration
import pwndbg.lib.config
import pwndbg.lib.tempfile
from pwndbg.commands import CommandCategory

# ========= Version / Installation code =========

# Bump me if needed.
# (This will trigger check_decomp2dbg_version_bumped())
d2d_required_major, d2d_required_minor, d2d_required_fix = 3, 14, 0

d2d_required_version_str: str = f"{d2d_required_major}.{d2d_required_minor}.{d2d_required_fix}"

d2d_cache_dir: Path = Path(pwndbg.lib.tempfile.cachedir("d2d"))


def decomp2dbg_not_installed_message() -> None:
    print(message.error("decomp2dbg is not installed.\n"))
    print(f"The supported decomp2dbg version is {d2d_required_major}.{d2d_required_minor}.* .")
    print(
        "If you installed Pwndbg with your distribution's package manager, use the same to install decomp2dbg."
    )
    print(
        "(If the version of decomp2dbg in your distro is outdated, complain to them. If our version is outdated, complain to us.)"
    )
    print(
        "If you're using Pwndbg installed from source, run `uv sync --extra decomp2dbg --inexact`."
    )


def decomp2dbg_path() -> Path:
    """
    Returns the absolute path to the directory where decomp2dbg is installed.

    If Pwndbg is installed from source this will be
    /path/to/pwndbg/.venv/lib/python3.13/site-packages/decomp2dbg.
    """
    import decomp2dbg

    return Path(decomp2dbg.__file__).parent.resolve()


def check_decomp2dbg_version() -> bool:
    """
    Checks the version of the decomp2dbg python module.

    Returns True if the version is supported, prints an error message
    and returns False otherwise.
    """
    import decomp2dbg

    ver_arr: list[str] = decomp2dbg.__version__.split(".")
    major: int = int(ver_arr[0])
    minor: int = int(ver_arr[1])
    fix: int = int(ver_arr[2])

    if major == d2d_required_major and (
        (minor > d2d_required_minor) or (minor == d2d_required_minor and fix >= d2d_required_fix)
    ):
        return True

    print(message.system("Unsupported decomp2dbg version installed."))
    msg = f"""
You have version {decomp2dbg.__version__} installed, but we need {d2d_required_version_str}.
This should only be possible if you installed Pwndbg through a package manager. You have a few options, in recommended order:

1. Complain to your distribution's packagers that this version of decomp2dbg is incompatible with this version of Pwndbg.
2. Install Pwndbg any other way: https://pwndbg.re/stable/setup/
3. Install the correct version of decomp2dbg manually.
"""
    print(msg)
    return False


def check_outdated_ghidra_plugin() -> bool:
    """
    Check if the user likely has an outdated version of the ghidra plugin.

    Tell them to update if yes and return False. If everything is up to date return True.

    This should only be checked if (decompiler_host == "localhost").
    """
    version_path: Path = d2d_cache_dir / "ghidra_plugin_version"
    if not version_path.exists():
        # The user has not installed the ghidra plugin yet.
        print(message.error("Ghidra plugin not installed."))
        print(f"(The file {version_path} does not exist)")
        return False

    # Note that we only save the "required version" to this file, because thats the only
    # thing that matters to us.
    version_spec = version_path.read_text().strip()
    if version_spec == d2d_required_version_str:
        # The version is the same i.e. we haven't bumped the required version since the
        # user installed the plugin with `di install ghidra`.
        return True

    # The plugin is outdated.
    print(
        message.error("Ghidra plugin outdated. ")
        + f"You have version {version_spec} but required is {d2d_required_version_str}."
    )
    print("Please run `di install ghidra`.")
    return False


def ghidra_decomp2dbg_version() -> None:
    """
    Save the required decomp2dbg version to a known file.
    """
    version_path: Path = d2d_cache_dir / "version"

    just_in_case: str = version_path.read_text().strip()
    if just_in_case == "fine":
        return

    version_path.write_text(d2d_required_version_str)


def print_d2d_version() -> None:
    import decomp2dbg

    print(f"decomp2dbg version: {decomp2dbg.__version__} (required: {d2d_required_version_str})")


ida_plugin_path = pwndbg.config.add_param(
    "decompiler-ida-plugin-path",
    str(Path.home() / ".idapro/plugins"),
    "where to install the ida integration plugin",
    param_class=pwndbg.lib.config.PARAM_STRING,
)
binja_plugin_path = pwndbg.config.add_param(
    "decompiler-binja-plugin-path",
    str(Path.home() / ".binaryninja/plugins"),
    "where to install the binary ninja integration plugin",
    param_class=pwndbg.lib.config.PARAM_STRING,
)
angr_plugin_path = pwndbg.config.add_param(
    "decompiler-angr-plugin-path",
    str(Path.home() / ".local/share/angr-management/plugins"),
    "where to install the angr integration plugin",
    param_class=pwndbg.lib.config.PARAM_STRING,
)


def install_generic_plugin(
    paths: list[tuple[Path, Path]],
    decomp_name: str,
    packaged_plugin_path: Path,
    config_var: pwndbg.lib.config.Parameter,
):
    """
    Arguments:
        paths: A list of (source path, destination path) tuples. Each element of the list
               will be symlinked (destination) -> (source).
        decomp_name: Pretty name of the decompiler.
        packaged_plugin_path: The path of the folder for the decompiler plugin in the decomp2dbg python package.
        config_var: The variable which holds the destination plugin path.
    """
    print(f"Installing the {decomp_name} decompiler plugin.\n")
    print_d2d_version()

    plugin_destination: Path = Path(str(config_var))

    print("\nSource:      ", packaged_plugin_path)
    print("Destination: ", plugin_destination)

    print("\nMaking sure destination folder exists..\n")
    plugin_destination.mkdir(parents=True, exist_ok=True)

    print("Deleting old files (if they exist):")
    for _, dest in paths:
        print(f"\t{dest}")
        if dest.exists() or dest.is_symlink():
            if dest.is_symlink() or not dest.is_dir():
                # Works for regular files and symlinks.
                # Note that `dest.is_dir()` returns True for
                # symlinks to directories.
                dest.unlink()
            else:
                # This is a non-symlink directory
                shutil.rmtree(str(dest))

    print("\nCreating symlinks:")
    for source, dest in paths:
        print(f"\t{dest} -> {source}")
        dest.symlink_to(source)

    print("\nThe fact that symlinks are used means the decompiler plugin will be automatically")
    print(
        f"updated when the decomp2dbg python package is updated. But {message.notice('if the decomp2dbg')}"
    )
    print(
        f"{message.notice('installation path gets changed')}, don't forget to {message.notice('reinstall')}!"
    )
    print("(so take care if you change the folder of your Pwndbg installation)\n")

    print(
        message.hint(
            f"If you want to change the plugin destination, run `set {config_var.name} the/new/path`.\n"
            "(and put this line into your ~/.gdbinit so you don't have issues in the future)\n"
        )
    )
    print(
        message.success("Installed successfully.")
        + " If your decompiler is already open, restart it. You can use `di connect` now."
    )


def install_ida_plugin() -> None:
    packaged_plugin_path: Path = decomp2dbg_path() / "decompilers/d2d_ida"
    plugin_destination: Path = Path(str(ida_plugin_path))

    packaged1 = packaged_plugin_path / "d2d_ida"
    packaged2 = packaged_plugin_path / "d2d_ida.py"

    dest1 = plugin_destination / "d2d_ida"
    dest2 = plugin_destination / "d2d_ida.py"

    install_generic_plugin(
        [(packaged1, dest1), (packaged2, dest2)], "IDA", packaged_plugin_path, ida_plugin_path
    )


def install_binja_plugin() -> None:
    packaged_plugin_path: Path = decomp2dbg_path() / "decompilers/d2d_binja"

    install_generic_plugin(
        [(packaged_plugin_path, Path(str(binja_plugin_path)) / "d2d_binja")],
        "Binary Ninja",
        packaged_plugin_path,
        binja_plugin_path,
    )


def install_angr_plugin() -> None:
    packaged_plugin_path: Path = decomp2dbg_path() / "decompilers/d2d_angr"

    install_generic_plugin(
        [(packaged_plugin_path, Path(str(angr_plugin_path)) / "d2d_angr")],
        "angr-managment",
        packaged_plugin_path,
        angr_plugin_path,
    )


def install_ghidra_plugin() -> None:
    print("Installing the Ghidra decompiler plugin.")
    print_d2d_version()

    download_url: str = f"https://github.com/mahaloz/decomp2dbg/releases/download/v{d2d_required_version_str}/d2d-ghidra-plugin.zip"
    download_dest: Path = d2d_cache_dir / "d2d-ghidra-plugin.zip"

    print("\nSince the Ghidra extension is written in Java, we download it as already built.")
    print(f"Downloading:\n\t{download_url}\n\t-> {download_dest}")

    try:
        with (
            urllib.request.urlopen(download_url) as response,
            open(str(download_dest), "wb") as out_file,
        ):
            shutil.copyfileobj(response, out_file)
    except urllib.error.HTTPError as e:
        print(message.error("\nHTTP Error while fetching the plugin. Aborting."))
        print("Status code:", e.code)
        print("Reason:", e.reason)
        print("Response body:", e.read())
        return
    except urllib.error.URLError as e:
        print(message.error("\nURL Error while fetching the plugin. Aborting."))
        print("Reason:", e.reason)
        return

    print(message.success("Done.\n"))

    print("Unfortunately, Ghidra doesn't load the plugin instantly on startup, so you ")
    print(
        message.notice("need to tell Ghidra to load the plugin")
        + " by clicking [File > Install Extensions > + (top right)]"
    )
    print(
        "in the Project Managment window. Then restart Ghidra. And in your project you might also need to"
    )
    print(
        "[File > Configure] and enable decomp2dbg decompiler server'. Now you can start the server with "
    )
    print("Ctrl+Shift+D as usual.")

    print(
        message.warn("\nIMPORTANT: ")
        + "Because the Ghidra plugin is not shipped compiled in the decomp2dbg python package, there is no symlink"
    )
    print("and " + message.warn("the plugin will not be automatically updated.\n"))

    version_path: Path = d2d_cache_dir / "ghidra_plugin_version"
    version_path.write_text(d2d_required_version_str)
    print(f"Saved current required version ({d2d_required_version_str}) (to {version_path}).")


def install(which_decompiler: str) -> None:
    if sys.platform == "win32":
        print(
            message.system(
                "Installation on Windows not supported. Install separately: https://github.com/mahaloz/decomp2dbg?tab=readme-ov-file#install ."
            )
        )
        return

    if str(decompiler_host) != "localhost":
        print(message.warn("decompiler-host != localhost: Why are you installing locally then?\n"))
        print("If your decompiler is on another machine, you also need to install the decompiler")
        print("plugin on that machine. Ideally you should do that using `di install` there, but")
        print("if that's not possible (e.g. if you're using Windows there), install the correct")
        print(
            f"version ({d2d_required_version_str}) of decomp2dbg on its own: https://github.com/mahaloz/decomp2dbg?tab=readme-ov-file#install .\n"
        )

        check = input("Are you sure you want to continue [y/N]: ")
        if check.lower() != "y":
            return

        print()

    # If the user is not connecting to localhost, but they decided to continue regardless,
    # then we will still enforce this check.
    if not check_decomp2dbg_version():
        return

    match which_decompiler:
        case "ida":
            install_ida_plugin()
        case "binja":
            install_binja_plugin()
        case "ghidra":
            install_ghidra_plugin()
        case "angr":
            install_angr_plugin()


# ========= End of Version / Installation code =========
# ========= decompiler-integration command handling =========

decompiler_host = pwndbg.config.add_param(
    "decompiler-host",
    "localhost",
    "the host where the decompiler is exposed",
    param_class=pwndbg.lib.config.PARAM_STRING,
)

decompiler_port = pwndbg.config.add_param(
    "decompiler-port",
    3662,
    "the port on which the decompiler is exposed",
    param_class=pwndbg.lib.config.PARAM_UINTEGER,
)


def disconnect() -> None:
    if not pwndbg.dintegration.manager.is_connected():
        print(message.error("Am not connected in the first place."))
        return

    decomp_name = pwndbg.dintegration.manager.decompiler_name()
    pwndbg.dintegration.manager.disconnect()
    print(message.success("Disconnected") + f" from {decomp_name}.")


def connect(also_sync: bool) -> None:
    # Doesn't make sense to check the version this if the local decomp2dbg is not being used.
    if decompiler_host == "localhost" and not check_decomp2dbg_version():
        return

    if pwndbg.dintegration.manager.is_connected():
        print("Reconnecting: ", end="")

    print(f"Connecting to {decompiler_host}:{decompiler_port}.")

    ok = pwndbg.dintegration.manager.connect(str(decompiler_host), int(decompiler_port))
    if ok:
        if decompiler_host != "localhost":
            print(
                "\nConnecting to a remote machine. "
                + message.system("Make sure")
                + " that the version"
            )
            print(
                f"of the decompiler plugin there is {message.system(d2d_required_version_str)}!\n"
            )
        else:
            # If we are connected to localhost Ghidra, we need to check that the plugin version is fine.
            if (
                pwndbg.dintegration.manager.decompiler_id()
                == pwndbg.dintegration.DecompilerID.GHIDRA
                and not check_outdated_ghidra_plugin()
            ):
                print(message.error("Disconnecting.."))
                pwndbg.dintegration.manager.disconnect()
                return

        decomp_name = pwndbg.dintegration.manager.decompiler_name()
        print(message.success("Connected") + f" to {decomp_name}.")

        if also_sync:
            # In case the binary isn't loaded yet, lets not yell to the user about failing.
            sync(fail_quietly=True)

        return

    print(message.error("Failed connecting."))
    print(message.hint("Did you open the connection in the decompiler? (Ctrl+Shift+D)"))
    print(
        message.hint(
            "(The appropriate decompiler plugin must be installed, see `di install --help`)"
        )
    )


def soft_connection_check(also_sync: bool) -> bool:
    """
    If we are not connected, try to connect (and sync).

    If we were connected, or succeed in connecting, return True,
    otherwise False.
    """
    if not pwndbg.dintegration.manager.is_connected():
        print("Trying to connect.. ", end="")

        connect(also_sync=also_sync)

        # Make sure we were successful.
        if not pwndbg.dintegration.manager.is_connected():
            return False

        # Give space to the actual command output.
        print()

    return True


def check_alive(error_msg: str) -> bool:
    try:
        inf = pwndbg.dbg.selected_inferior()

        if not inf.alive():
            # A bit hacky but whatever.
            raise pwndbg.dbg_mod.NoInferior

        return True
    except pwndbg.dbg_mod.NoInferior:
        print(message.error(error_msg))
        return False


def jump(addr: int | None) -> None:
    if not pwndbg.dintegration.manager.is_connected():
        print(message.error("Not connected to a decompiler."))
        print(message.hint("Try `di connect`."))
        return

    if not check_alive("Can only jump to address while the process is alive."):
        return

    if addr is None:
        if pwndbg.aglib.regs.pc is None:
            print(message.error("Address not specified, and could not find PC."))
            return
        addr = pwndbg.aglib.regs.pc

    ok = pwndbg.dintegration.manager.focus_address(addr)
    if not ok:
        print(message.error("Decompiler failed to jump."))


def sync(fail_quietly: bool) -> None:
    """
    Arguments:
        fail_quietly: If we don't pass the preliminary checks required to perform the sync, don't print anything.
    """
    if fail_quietly:
        # Direct check, no retries.
        if not pwndbg.dintegration.manager.is_connected():
            return

        # Something else is calling us, lets give the output some space.
        print()
    else:
        # Noisy check with a connection attempt.
        # Don't try to sync because that sync would be quiet, and we want
        # to complain about errors to the user.
        if not soft_connection_check(also_sync=False):
            return

    if not check_alive(
        "" if fail_quietly else "Can only sync with the debugger while the process is alive."
    ):
        return

    print("Syncing symbols...")

    # Functions and globals
    nsyms, sym_err = pwndbg.dintegration.manager.update_symbols()
    match sym_err:
        case pwndbg.dintegration.Error.OK:
            if nsyms == 0:
                print("No symbols synced? Something is off. ")
            else:
                print(
                    message.success(f"Synced {nsyms} symbols") + " (globals + functions). ", end=""
                )
        case pwndbg.dintegration.Error.DEBUGGER_NOT_SUPPORTED:
            print("LLDB does not support syncing symbols. ", end="")
        case _:
            print(message.error(f"Failed: {sym_err.value}."))
            if sym_err == pwndbg.dintegration.Error.BINARY_NOT_LOADED:
                print(message.hint("Try `di setpath --help` or `di setbase --help`?"))
            # The error is fundamental to the setup, don't even try to sync function variables.
            return

    # Function-local variables
    nvars, var_err = pwndbg.dintegration.manager.update_function_variables()
    match var_err:
        case pwndbg.dintegration.Error.OK:
            if nvars > 0:
                print(message.success(f"Synced {nvars} variables") + " for the current function.")
            else:
                # It's fine to print this even if fail_quietly=True.
                print("No variables synced for the current function.")
        case pwndbg.dintegration.Error.NO_FRAME:
            # It's fine to print this even if fail_quietly=True.
            print("No variables synced for the current function (no stack frame found).")
        case pwndbg.dintegration.Error.NO_CONNECTION:
            print(message.error(f"Failed: {sym_err.value}."))


def list_one_frame(frame: pwndbg.dbg_mod.Frame, idx: int | None = None) -> None:
    func_vars: pwndbg.dintegration.RebasedFuncVariables | None = (
        pwndbg.dintegration.manager.get_function_vars_rebased_from_frame(frame)
    )

    pc: int = frame.pc()
    sp: int = frame.sp()
    start: int | None = frame.start()

    symbol: str | None = pwndbg.aglib.symbol.resolve_addr(pc)
    if symbol:
        symbol_text = color.blue(symbol)
    else:
        symbol_text = "???"

    if idx is not None:
        frame_text = f"#{idx} {symbol_text} frame:"
    else:
        frame_text = f"{symbol_text} frame:"

    pc_text = color.blue(hex(pc))
    sp_text = color_mem.get(sp)
    start_text = color_mem.get(start) if start is not None else "???"
    padding = " " * 4

    print(frame_text)
    print(f"{padding}@ {pc_text}")
    print(f"{padding}{sp_text} -> {start_text}")

    if func_vars is None:
        # Common reason is that we are in a function in a different binary.
        print("Could not get function variables from decompiler.")
        return

    if len(func_vars.reg_vars) == 0:
        print("No register variables.")
    else:
        print("Register variables:")

        for reg_var in func_vars.reg_vars:
            name_text = color.green(color.bold(reg_var.name))
            type_text = color.light_cyan(reg_var.type)
            reg_text = reg_var.reg_name.ljust(4, " ")
            # FIXME: Should probably refactor this to use pwndbg.commands.context.get_regs (but then also
            # refactor that, to pull it out of pwndbg/commands, maybe separate out register name and value etc.)
            reg_value_raw: pwndbg.dbg_mod.Value | None = frame.regs().by_name(reg_var.reg_name)
            try:
                reg_value = (
                    color_mem.get(int(reg_value_raw))
                    if reg_value_raw is not None
                    else color.gray("???")
                )
            except pwndbg.dbg_mod.Error:
                # int(reg_value_raw) failed. Happens for xmm0 for instance.
                reg_value = "not an int"
            reg_value_part = color.ljust_colored(f"(value: {reg_value})", 28)
            print(f"{reg_text} {reg_value_part} <- {name_text} (type: {type_text})")

    if len(func_vars.stack_vars) == 0:
        print("No stack variables.")
    else:
        print("Stack variables:")

        for stack_var in func_vars.stack_vars:
            name_text = color.green(color.bold(stack_var.name))
            type_text = color.light_cyan(stack_var.type)
            addr_text = color.ljust_colored(color_mem.get(stack_var.addr), 18)
            from_sp = stack_var.addr - sp
            from_sp_text = f"[sp + {from_sp:#x}]"
            if start:
                from_frame = start - stack_var.addr
                from_frame_text = f"[frame - {from_frame:#x}]"
            else:
                from_frame_text = "[???]"

            print(
                f"{addr_text} <- {name_text} (type: {type_text}) {from_sp_text} {from_frame_text}"
            )


def list_all_frames() -> None:
    thread = pwndbg.dbg.selected_thread()
    if thread is None:
        print(message.error("Could not find current thread."))
        return

    idx = 0
    with thread.bottom_frame() as bottom_frame:
        cur_frame = bottom_frame
        # Crawl up the stack
        while cur_frame is not None:
            list_one_frame(cur_frame, idx)
            print("==================")
            cur_frame = cur_frame.parent()
            idx += 1


def list_(list_all: bool) -> None:
    if not soft_connection_check(also_sync=True):
        return

    if not check_alive("Can only list function variables if the process is alive."):
        return

    if list_all:
        list_all_frames()
    else:
        frame: pwndbg.dbg_mod.Frame | None = pwndbg.dbg.selected_frame()
        if frame is None:
            print(message.error("Could not find current stack frame."))
            return
        list_one_frame(frame)


def setpath(path: str) -> None:
    # I make this a command instead of a config for consistency with setbase.

    # Unset manual base first
    if pwndbg.dintegration.manual_binary_address != -1:
        print(
            f"Unset the previously set `di setbase` value of {pwndbg.dintegration.manual_binary_address}."
        )
        pwndbg.dintegration.manual_binary_address = -1

    pwndbg.dintegration.manual_binary_path = path
    print(f'Path of the decompiled binary in the address space set to "{path}".')
    if path == "":
        print("(back to automatic detection)")

    if pwndbg.dintegration.manager.is_connected():
        print("Reconnecting to apply changes..\n")
        connect(also_sync=True)


def setbase(base_addr: int) -> None:
    # I use a command like this instead of a config parameter because it seems
    # GDB doesn't allow values > 2^32.
    if base_addr < -1:
        print(message.error("Valid values are in [-1, 2^64)."))
        return

    # Unset manual path first
    if pwndbg.dintegration.manual_binary_path != "":
        print(
            f"Unset the previously set `di setpath` value of {pwndbg.dintegration.manual_binary_path}."
        )
        pwndbg.dintegration.manual_binary_path = ""

    pwndbg.dintegration.manual_binary_address = base_addr
    print(f"Base address of the decompiled binary set to {base_addr:#x}.")
    if base_addr == -1:
        print("(back to automatic detection)")

    if pwndbg.dintegration.manager.is_connected():
        print("Reconnecting to apply changes..\n")
        connect(also_sync=True)


parser = argparse.ArgumentParser(
    description="""Control Pwndbg decompiler integration.

See https://pwndbg.re/dev/tutorials/decompiler-integration/ for usage instructions."""
)
subparsers = parser.add_subparsers(dest="command")
subparsers.required = True

parser_connect = subparsers.add_parser(
    "connect",
    aliases=["c"],
    help="Connect to the decompiler",
    description="""
Connect to the decompiler.

The host and port to connect to are governed by the `decompiler-host`
and `decompiler-port` config variables. Try `help set decompiler-host`.
""",
)

parser_disconnect = subparsers.add_parser(
    "disconnect",
    aliases=["d"],
    help="Disconnect from the decompiler",
    description="Disconnect from the decompiler.",
)

parser_sync = subparsers.add_parser(
    "sync",
    aliases=["s"],
    help="Sync data from the decompiler",
    description="""
Sync data from the decompiler.

Check out `help set decompiler-autosync-syms` and `help set decompiler-autosync-vars`.
""",
)

parser_jump = subparsers.add_parser(
    "jump",
    aliases=["j"],
    help="Make the decompiler's cursor jump to the PC",
    description="""
Make the decompiler's cursor jump to the PC.

Check out `help set decompiler-autojump`.
""",
)
parser_jump.add_argument(
    "jump_addr",
    metavar="addr",
    type=int,
    nargs="?",
    default=None,
    help="Address to jump to. (default: pc)",
)

parser_install = subparsers.add_parser(
    "install",
    help="Install the decompiler plugins",
    description="""
Install/update the decompiler plugins.

You need a decompiler plugin installed to allow the decompiler to communicate
back to Pwndbg. The decompiler plugins are from decomp2dbg (<3).

If you already have decomp2dbg installed, this command will overwrite
that installation in order to pin the proper version that Pwndbg needs. You will
still be able to use decomp2dbg outside of Pwndbg.

You should take care not to invoke `source /path/to/decomp2dbg/d2d.py` in your ~/.gdbinit
because we implement the debugger-side logic independently, and it might conflict.
""",
)
install_subparsers = parser_install.add_subparsers(dest="install_sub", metavar="which")
install_subparsers.required = True

parser_install_ida = install_subparsers.add_parser(
    "ida",
    help="Install the IDA decompiler plugin",
    description="Install the IDA decompiler plugin.",
)
parser_install_binja = install_subparsers.add_parser(
    "binja",
    help="Install the Binary Ninja decompiler plugin",
    description="Install the Binary Ninja decompiler plugin.",
)
parser_install_ghidra = install_subparsers.add_parser(
    "ghidra",
    help="Install the Ghidra decompiler plugin",
    description="Install the Ghidra decompiler plugin.",
)
parser_install_angr = install_subparsers.add_parser(
    "angr",
    help="Install the angr-management decompiler plugin",
    description="Install the angr-managment decompiler plugin.",
)

parser_decomp = subparsers.add_parser(
    "decomp",
    help="Just use the `decomp` command",
    description="Just use the `decomp` command.",
)

parser_list = subparsers.add_parser(
    "list",
    aliases=["l"],
    help="List the variables for the current stack frame",
    description="""
List the variables for the current stack frame.

Will not be accurate in a function's prologue (before the stack pointer has been adjusted).
The "frame" for the purposes of this command is (usually) the location of the saved return address.
""",
)
parser_list.add_argument(
    "-a",
    "--all",
    help="List decompiler stack variables from all stack frames in this thread.",
    action="store_true",
    default=False,
    dest="list_all",
)

parser_set_base = subparsers.add_parser(
    "setbase",
    help="Manually set the base memory address of the decompiled binary",
    description="""
Manually set the base memory address of the decompiled binary.

Normally, Pwndbg will use the file path that the decompiler reports for the binary and
check it against all files mapped into memory to find the correct base address.

If for some reason the file names differ or your binary does not show up in the memory
mappings, you can manually set the base address using this command. This is commonly
needed when debugging a kernel module.

If you wish to re-enable automatic base address detection, set this value to -1 (or
restart Pwndbg).

Setting this automatically unsets the `di setpath` value.
""",
)
parser_set_base.add_argument(
    "binary_addr",
    metavar="addr",
    type=int,
    help="Memory address of the decompiled binary in the address space",
)

parser_set_path = subparsers.add_parser(
    "setpath",
    help="Manually set the path of the binary as loaded in memory",
    description="""
Manually set the path of the binary as loaded in memory.

Normally, Pwndbg will use the file path that the decompiler reports for the binary and
check it against all files mapped into memory to find the correct base address.

If for some reason the file names differ or your binary does not show up in the memory
mappings, you can manually specify the actual path of the binary as loaded in memory (
the one reported by /proc/<pid>/maps i.e. vmmap).

If you wish to re-enable automatic base address detection, set this value to "" (or
restart Pwndbg).

Setting this automatically unsets the `di setbase` value.
""",
)
parser_set_path.add_argument(
    "binary_path",
    metavar="path",
    type=str,
    help="File path of the decompiled binary as loaded in memory",
)


@pwndbg.commands.Command(
    parser, aliases=["di"], category=pwndbg.commands.CommandCategory.INTEGRATIONS
)
def decompiler_integration(
    command: str,
    jump_addr: int | None = None,
    install_sub: str = "",
    list_all: bool = False,
    binary_addr: int = -1,
    binary_path: str = "",
):
    # decomp2dbg is an optional dependancy for now, so we check for it.
    try:
        match command:
            case "connect" | "c":
                connect(also_sync=True)
            case "disconnect" | "d":
                disconnect()
            case "sync" | "s":
                sync(fail_quietly=False)
            case "jump" | "j":
                jump(jump_addr)
            case "install":
                install(install_sub)
            case "decomp":
                print(message.notice("Just use the `decomp` command."))
            case "list" | "l":
                list_(list_all)
            case "setbase":
                setbase(binary_addr)
            case "setpath":
                setpath(binary_path)
    except ModuleNotFoundError as e:
        if e.name != "decomp2dbg":
            raise e
        decomp2dbg_not_installed_message()


# ========= End of decompiler-integration command handling =========
# ========= Automatic integration handling =========

should_autosync_syms = pwndbg.config.add_param(
    "decompiler-autosync-syms",
    False,
    "whether to sync symbols with the decompiler on every stop",
    param_class=pwndbg.lib.config.PARAM_BOOLEAN,
    help_docstring="""
Depending on the decompiler, the number of symbols (functions + global variables)
the binary you are decompiling has, and various other factors, this may or may not
be a good idea. Try it out and see.

Check out the other decompiler-auto* configuration variables as well.
""",
)

should_autosync_vars = pwndbg.config.add_param(
    "decompiler-autosync-vars",
    True,
    "whether to sync function variables with the decompiler on every stop",
    param_class=pwndbg.lib.config.PARAM_BOOLEAN,
    help_docstring="""
This is generally lightweight, so it is enabled by default. Try disabling
it if you have performance issues.

Check out the other decompiler-auto* configuration variables as well.
""",
)

should_autojump = pwndbg.config.add_param(
    "decompiler-autojump",
    False,
    "whether to jump the decompiler cursor on every stop",
    param_class=pwndbg.lib.config.PARAM_BOOLEAN,
    help_docstring="""
Depending on the decompiler, this may or may not be a good idea.
Try it out and see.

Check out the other decompiler-auto* configuration variables as well.
""",
)


def auto_jump():
    if pwndbg.aglib.regs.pc is None:
        return
    addr: int = pwndbg.aglib.regs.pc

    pwndbg.dintegration.manager.focus_address(addr)


@pwndbg.dbg.event_handler(pwndbg.dbg_mod.EventType.STOP)
def automatic_operations() -> None:
    # The connection and inf.alive() checks in sync() are just for better error
    # reporting, the manager will handle them anyway.

    # We succeed quietly to not mess up the `context-reserve-lines` logic.

    if should_autosync_syms:
        pwndbg.dintegration.manager.update_symbols()

    if should_autosync_vars:
        pwndbg.dintegration.manager.update_function_variables()

    if should_autojump:
        auto_jump()


# ========= End of Automatic integration handling =========
# ========= The decomp command =========

parser = argparse.ArgumentParser(
    description="Use the current integration to decompile code near an address."
)

parser.add_argument(
    "addr",
    type=int,
    nargs="?",
    default=None,
    help="Address to decompile near. (default: pc)",
)
parser.add_argument(
    "lines",
    type=int,
    nargs="?",
    # Same as the default for context-code-lines
    default=14,
    help="Number of lines of decompilation to show.",
)


@pwndbg.commands.Command(parser, category=CommandCategory.INTEGRATIONS)
@pwndbg.commands.OnlyWhenRunning
def decomp(addr: int | None, lines: int) -> None:
    if addr is None:
        if pwndbg.aglib.regs.pc is None:
            print("Address not specified, and could not find PC.")
            return
        addr = pwndbg.aglib.regs.pc

    if not soft_connection_check(also_sync=True):
        return

    decomp = pwndbg.dintegration.manager.decompile_pretty(addr, lines)

    if decomp is None:
        print("Could not retrieve decompilation.")
    else:
        print("\n".join(decomp))



from __future__ import annotations

import argparse

import pwndbg.aglib.kernel
import pwndbg.commands
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(description="Return the kernel commandline (/proc/cmdline).")


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelDebugInfo
@pwndbg.commands.OnlyWhenPagingEnabled
def kcmdline() -> None:
    print(pwndbg.aglib.kernel.kcmdline())



from __future__ import annotations

import argparse

import pwndbg.aglib.memory
import pwndbg.chain
import pwndbg.color as color
import pwndbg.commands
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(description="Dumps the arguments of a va_list.")
parser.add_argument("addr", type=int, help="Address of the va_list")
parser.add_argument("count", type=int, nargs="?", default=8, help="Number of arguments to dump")


@pwndbg.commands.Command(parser, category=CommandCategory.MISC)
@pwndbg.commands.OnlyWhenRunning
def valist(addr: int, count: int) -> None:
    # The `va_list` struct looks like this:
    #
    # ```
    # typedef struct {
    #    unsigned int gp_offset;
    #    unsigned int fp_offset;
    #    void *overflow_arg_area;
    #    void *reg_save_area;
    # } va_list[1];
    # ```

    gp_offset = pwndbg.aglib.memory.u32(addr)
    gp_index = gp_offset / 8

    overflow_arg_area = pwndbg.aglib.memory.u64(addr + 8)
    reg_save_area = pwndbg.aglib.memory.u64(addr + 16)

    indent = " " * len("gp_offset => ")
    print(f"{color.blue('reg_save_area')}")
    for i in range(6):
        line = ""
        if i == gp_index:
            line += "gp_offset => "
        else:
            line += indent

        line += pwndbg.chain.format(reg_save_area + i * 8)
        print(line)

    print()
    print(f"{color.blue('overflow_arg_area')}")
    for i in range(count - 6):
        print(indent + pwndbg.chain.format(overflow_arg_area + i * 8))



from __future__ import annotations

import gdb

from pwndbg.gdblib.tui import context as context
from pwndbg.gdblib.tui import control as control

# Run side-effects^^


def setup() -> None:
    tui_layouts = [
        (
            "tui new-layout pwndbg "
            "{-horizontal "
            " { "
            "  { -horizontal "
            "   { pwndbg_disasm 1 } 2 "
            "   { "
            "     { -horizontal pwndbg_legend 8 pwndbg_control 2 } 0 pwndbg_regs 1 pwndbg_stack 1 "
            "   } 3 "
            "  } 7 cmd 3 "
            " } 3 { pwndbg_backtrace 2 pwndbg_expressions 2 pwndbg_threads 1 } 1 "
            "} 1 status 1"
        ),
        (
            "tui new-layout pwndbg_code "
            "{-horizontal "
            " { "
            "  { -horizontal "
            "   { pwndbg_code 1 pwndbg_disasm 1 } 2 "
            "   { "
            "     { -horizontal pwndbg_legend 8 pwndbg_control 2 } 0 pwndbg_regs 1 pwndbg_stack 1 "
            "   } 3 "
            "  } 7 cmd 3 "
            " } 3 { pwndbg_backtrace 2 pwndbg_expressions 2 pwndbg_threads 1 } 1 "
            "} 1 status 1"
        ),
    ]
    for layout in tui_layouts:
        try:
            gdb.execute(layout)
        except gdb.error:
            pass



"""
Command to print the information of the current Thread Local Storage (TLS).
"""

from __future__ import annotations

import argparse

from tabulate import tabulate

import pwndbg.aglib.dt
import pwndbg.aglib.memory
import pwndbg.aglib.symbol
import pwndbg.aglib.tls
import pwndbg.aglib.vmmap
import pwndbg.color
import pwndbg.color.memory as mem_color
import pwndbg.commands
import pwndbg.commands.context
import pwndbg.commands.telescope
from pwndbg.color import message
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="Print out base address of the current Thread Local Storage (TLS).",
)

parser.add_argument(
    "-p",
    "--pthread-self",
    action="store_true",
    default=False,
    help="Try to get the address of TLS by calling pthread_self().",
)

parser.add_argument("-a", "--all", action="store_true", help="Do not truncate the dump output.")


@pwndbg.commands.Command(parser, category=CommandCategory.LINUX)
@pwndbg.commands.OnlyWhenRunning
@pwndbg.commands.OnlyWhenUserspace
def tls(pthread_self=False, all: bool = False) -> None:
    tls_base = (
        pwndbg.aglib.tls.find_address_with_register()
        if not pthread_self
        else pwndbg.aglib.tls.find_address_with_pthread_self()
    )
    if pwndbg.aglib.memory.is_readable_address(tls_base):
        print(message.success(f"Thread Local Storage (TLS) base: {tls_base:#x}"))
        print(message.success("TLS is located at:"))
        print(message.notice(pwndbg.aglib.vmmap.find(tls_base)))

        # Displaying `dt tcbhead_t <tls_base>` if possible
        # If not, we will dump the tls with telescope
        output = str(pwndbg.aglib.dt.dt("tcbhead_t", addr=tls_base))

        print(message.success("Dumping the address:"))
        if output == "Type not found.":
            pwndbg.commands.telescope.telescope(tls_base, 10)
        else:
            lines = output.splitlines()

            if all or len(lines) <= 10:
                print(message.notice(output))
            else:
                index = None
                for i, line in enumerate(lines):
                    if "__glibc_unused2" in line:
                        index = i
                        break

                if index is not None:
                    end_index = index + 2
                    for line in lines[:end_index]:
                        print(message.notice(line))
                    print(message.notice("\t[...]"))
                    print(
                        message.hint(
                            "Output truncated. Rerun with option -a to display the full output."
                        )
                    )
                # In case there is a tcbhead_t but there is no __glibc_unused2
                else:
                    for line in lines[:10]:
                        print(message.notice(line))
                    print(message.notice("\t[...]"))
                    print(
                        message.hint(
                            "Output truncated. Rerun with option -a to display the full output."
                        )
                    )
        return
    print(message.error("Couldn't find Thread Local Storage (TLS) base."))
    if not pthread_self:
        print(
            message.notice(
                "You can try to use -p/--pthread option to get the address of TLS by calling pthread_self().\n"
                "(This might cause problems if the pthread_self() is not in libc or not initialized yet.)"
            )
        )


parser = argparse.ArgumentParser(
    description="List all threads belonging to the selected inferior.",
)
group = parser.add_mutually_exclusive_group()

group.add_argument(
    "num_threads",
    type=int,
    nargs="?",
    default=None,
    help="Number of threads to display. Omit to display all threads.",
)

group.add_argument(
    "-c",
    "--config",
    action="store_true",
    dest="respect_config",
    help="Respect context-max-threads config to limit number of threads displayed.",
)


@pwndbg.commands.Command(parser, category=CommandCategory.LINUX)
@pwndbg.commands.OnlyWhenRunning
@pwndbg.commands.OnlyWhenUserspace
def threads(num_threads, respect_config) -> None:
    table = []
    headers = ["global_num", "name", "status", "pc", "symbol"]
    bold_green = lambda text: pwndbg.color.bold(pwndbg.color.green(text))

    import gdb

    try:
        original_thread = gdb.selected_thread()
    except SystemError:
        original_thread = None
    try:
        original_frame = gdb.selected_frame()
    except gdb.error:
        original_frame = None

    all_threads = gdb.selected_inferior().threads()[::-1]

    displayed_threads = []

    if original_thread is not None and original_thread.is_valid():
        displayed_threads.append(original_thread)

    for thread in all_threads:
        if respect_config and len(displayed_threads) >= int(
            pwndbg.commands.context.config_max_threads_display
        ):
            break

        if num_threads is not None and len(displayed_threads) >= num_threads:
            break

        if thread.is_valid() and thread is not original_thread:
            displayed_threads.append(thread)

    for thread in displayed_threads:
        name = thread.name or ""

        if thread is original_thread:
            row = [
                bold_green(thread.global_num),
                bold_green(name),
            ]
        else:
            row = [
                str(thread.global_num),
                name,
            ]

        row.append(pwndbg.commands.context.get_thread_status(thread))

        if thread.is_stopped():
            thread.switch()
            pc = pwndbg.dbg.selected_frame().pc()

            pc_colored = mem_color.get(pc)
            symbol = pwndbg.aglib.symbol.resolve_addr(pc)

            row.append(pc_colored)

            if symbol:
                if thread is original_thread:
                    row.append(bold_green(symbol))
                else:
                    row.append(symbol)

        table.append(row)

    if original_thread is not None and original_thread.is_valid():
        original_thread.switch()
    if original_frame is not None and original_frame.is_valid():
        original_frame.select()

    print(tabulate(table, headers))
    print(f"\nShowing {len(displayed_threads)} of {len(all_threads)} threads.")



"""
Launches the target process after setting a breakpoint at a convenient
entry point.
"""

from __future__ import annotations

import argparse
from shlex import quote

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.elf
import pwndbg.aglib.proc
import pwndbg.aglib.symbol
import pwndbg.color.message as message
import pwndbg.commands
import pwndbg.dbg_mod
from pwndbg.commands import CommandCategory
from pwndbg.dbg_mod import BreakpointLocation
from pwndbg.dbg_mod import DebuggerType

if pwndbg.dbg.is_gdblib_available():
    import gdb


def breakpoint_at_entry():
    addr = int(pwndbg.aglib.elf.entry())
    if not addr:
        print(message.error("No entry address found for the binary."))
        return

    if int(pwndbg.aglib.regs.pc) == addr:
        # Skip setting the breakpoint because we are already at the entry point.
        # This occurs when execution started with `starti` or `run -s`.
        return

    proc = pwndbg.dbg.selected_inferior()
    bp = proc.break_at(BreakpointLocation(addr), internal=True)

    async def ctrl(ec: pwndbg.dbg_mod.ExecutionController):
        with bp:
            await ec.cont(bp)

    proc.dispatch_execution_controller(ctrl)


# Starting from 3rd paragraph, the description is
# taken from the GDB's `starti` command description
parser = argparse.ArgumentParser(
    description="""
Start the debugged program stopping at the first convenient location
from this list: main, _main, start, _start, init or _init.
You may specify arguments to give it.

Args may include "*", or "[...]"; they are expanded using the
shell that will start the program (specified by the "$SHELL" environment
variable).  Input and output redirection with ">", "<", or ">>"
are also allowed.

With no arguments, uses arguments last specified (with "run" or
"set args").  To cancel previous arguments and run with no arguments,
use "set args" without arguments.

To start the inferior without using a shell, use "set startup-with-shell off".
""",
)

parser.add_argument(
    "args", nargs="*", type=str, default=None, help="The arguments to run the binary with."
)


@pwndbg.commands.Command(
    parser,
    aliases=["main", "init"],
    only_debuggers={DebuggerType.GDB},
    category=CommandCategory.START,
)
@pwndbg.commands.OnlyWhenLocal
def start(args=None) -> None:
    if args is None:
        args = []
    run = "run " + " ".join(args)

    symbols = ["main", "_main", "start", "_start", "init", "_init"]

    for symbol in symbols:
        address = pwndbg.aglib.symbol.lookup_symbol_addr(symbol)
        if not address:
            continue

        gdb.Breakpoint(symbol, temporary=True)
        gdb.execute(run, from_tty=False)
        return

    # Try a breakpoint at the binary entry
    entry(args)


# Starting from 3rd paragraph, the description is
# taken from the GDB's `starti` command description
parser = argparse.ArgumentParser(
    description="""
Start the debugged program stopping at its entrypoint address.

Note that the entrypoint may not be the first instruction executed
by the program. If you want to stop on the first executed instruction,
use the GDB's `starti` command or LLDB's `process launch -s`.

Args may include "*", or "[...]"; they are expanded using the
shell that will start the program (specified by the "$SHELL" environment
variable).  Input and output redirection with ">", "<", or ">>"
are also allowed.

With no arguments, uses arguments last specified (with "run" or
"set args").  To cancel previous arguments and run with no arguments,
use "set args" without arguments.

To start the inferior without using a shell, use "set startup-with-shell off".
""",
)
parser.add_argument(
    "args", nargs="*", type=str, default=None, help="The arguments to run the binary with."
)


@pwndbg.commands.Command(parser, category=CommandCategory.START)
@pwndbg.commands.OnlyWithFile
@pwndbg.commands.OnlyWhenLocal
def entry(args=None) -> None:
    if args is None:
        args = []

    if pwndbg.dbg.is_gdblib_available():
        # If this is GDB, just start the process ourselves.
        run = "starti " + " ".join(map(quote, args))
        gdb.execute(run, from_tty=False)
    else:
        # For now, there is no debugger-agnostic way to start a process from
        # inside a command, so the best we can do is expect that the back-end
        # picks up that this command is being called, and starts the process on
        # our behalf, and error out if it does not.
        #
        # `pwndbg-lldb` implements starting as a partial command override in the CLI.
        #
        # TODO: In the future, we should handle starts using an in-command mechanism.
        if not pwndbg.aglib.proc.alive():
            print(
                message.error(
                    "The program is not running. Start the program with `run -s` and then use `entry` to set the breakpoint."
                )
            )
            return
    breakpoint_at_entry()


@pwndbg.commands.Command(
    "Alias for 'tbreak __libc_start_main; run'.",
    only_debuggers={DebuggerType.GDB},
    category=CommandCategory.START,
)
@pwndbg.commands.OnlyWithFile
@pwndbg.commands.OnlyWhenLocal
def sstart() -> None:
    gdb.Breakpoint("__libc_start_main", temporary=True)
    gdb.execute("run")



"""
Common helper and cache for pwndbg tempdir
"""

from __future__ import annotations

import os
import sys
import tempfile
from pathlib import Path

import pwndbg.lib.cache


@pwndbg.lib.cache.cache_until("forever")
def tempdir() -> Path:
    """
    Returns a safe and unpredictable temporary directory with pwndbg prefix.
    """
    return Path(tempfile.mkdtemp(prefix="pwndbg-"))


@pwndbg.lib.cache.cache_until("forever")
def cachedir(namespace: str | None = None) -> Path:
    """
    Returns and potentially creates a persistent safe cachedir location
    based on XDG_CACHE_HOME or ~/.cache or LOCALAPPDATA (Windows)

    Optionally creates a sub namespace inside the pwndbg cache folder.
    """
    if sys.platform == "win32":
        cachehome = os.getenv("LOCALAPPDATA")
    else:
        cachehome = os.getenv("XDG_CACHE_HOME") or os.path.join(os.getenv("HOME", ""), ".cache")
    cachedir = os.path.join(cachehome or tempfile.gettempdir(), "pwndbg")
    if namespace:
        cachedir = os.path.join(cachedir, namespace)
    os.makedirs(cachedir, exist_ok=True)
    return Path(cachedir)



from __future__ import annotations

from collections.abc import Callable
from typing import Any

import pwndbg.aglib.qemu
import pwndbg.aglib.stack
import pwndbg.aglib.symbol
import pwndbg.aglib.vmmap
import pwndbg.dbg_mod
import pwndbg.lib.memory
import pwndbg.lib.pretty_print
from pwndbg.color import ColorConfig
from pwndbg.color import ColorParamSpec
from pwndbg.color import normal

ColorFunction = Callable[[str], str]

c = ColorConfig(
    "memory",
    [
        ColorParamSpec("stack", "yellow", "color for stack memory"),
        ColorParamSpec("heap", "blue", "color for heap memory"),
        ColorParamSpec("code", "red", "color for executable memory"),
        ColorParamSpec("data", "purple", "color for all other writable memory"),
        ColorParamSpec("rodata", "normal", "color for all read only memory"),
        ColorParamSpec("wx", "underline", "color added to all WX memory"),
        ColorParamSpec("guard", "cyan", "color added to all guard pages (no perms)"),
    ],
)


def get_address_and_symbol(address: int, decompiler_stack_variables: dict[int, str]) -> str:
    """
    Convert and colorize address 0x7ffff7fcecd0 to string `0x7ffff7fcecd0 (_dl_fini)`
    If no symbol exists for the address, return colorized address
    """
    symbol = pwndbg.aglib.symbol.resolve_addr(address)
    if symbol:
        symbol = f"{address:#x} ({symbol})"
    else:
        var: str | None = pwndbg.aglib.stack.get_stack_var_name(address)
        if var is None:
            var = decompiler_stack_variables.get(address)
        if var is not None:
            symbol = f"{address:#x} {{{var}}}"
    return get(address, symbol)


def get_address_or_symbol(address: int, decompiler_stack_variables: dict[int, str]) -> str:
    """
    Convert and colorize address to symbol if it can be resolved, else return colorized address
    """
    return attempt_colorized_symbol(address, decompiler_stack_variables) or get(address)


def attempt_colorized_symbol(
    address: int, decompiler_stack_variables: dict[int, str]
) -> str | None:
    """
    Convert address to colorized symbol (if symbol is there), else None
    """
    symbol = pwndbg.aglib.symbol.resolve_addr(address)
    if symbol:
        return get(address, symbol)
    var: str | None = pwndbg.aglib.stack.get_stack_var_name(address)
    if var is None:
        var = decompiler_stack_variables.get(address)
    if var is not None:
        return get(address, f"{{{var}}}")
    return None


# We have to accept `Any` here, as users may pass gdb.Value objects to this
# function. This is probably more lenient than we'd really like.
#
# TODO: Remove the exception for gdb.Value case from `pwndbg.color.memory.get`.
def get(
    address: int | pwndbg.dbg_mod.Value | Any,
    text: str | None = None,
    prefix: str | None = None,
    page: pwndbg.lib.memory.Page | None = None,
) -> str:
    """
    Returns a colorized string representing the provided address.

    Arguments:
        address: Address to look up
        text: Optional text to use in place of the address in the return value string.
        prefix: Optional text to set at beginning in the return value string, followed by a space, without modifiying the original text.
    """
    address = int(address)
    if page is None:  # if we know the containing page, don't bother to find it as an optimization
        page = pwndbg.aglib.vmmap.find(address)

    color: Callable[[str], str]

    if page is None:
        color = normal
    elif "[stack" in page.objfile:
        color = c.stack
    elif page.execute:
        color = c.code
    elif not page.write:
        color = c.rodata
    elif "[heap" in page.objfile or (
        pwndbg.aglib.qemu.is_qemu_kernel()
        and any(keyword in page.objfile for keyword in ("physmap", "vmalloc", "slab virtual"))
    ):
        color = c.heap
    elif page.rw:
        color = c.data
    elif page.is_guard:
        color = c.guard
    else:
        color = c.rodata

    if page and page.wx:
        old_color = color
        color = lambda x: c.wx(old_color(x))

    if text is None:
        text = pwndbg.lib.pretty_print.int_to_string(address)

    if prefix is not None:
        # Prepend the prefix and a space before the existing text
        text = f"{prefix} {text}"

    return color(text)


def legend() -> str:
    return "LEGEND: " + " | ".join(
        (
            c.stack("STACK"),
            c.heap("HEAP"),
            c.code("CODE"),
            c.data("DATA"),
            # WX segments will also be marked as code, so do 2 formatters here
            c.wx(c.code("WX")),
            c.rodata("RODATA"),
        )
    )



from __future__ import annotations

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.memory
import pwndbg.aglib.typeinfo
import pwndbg.dbg_mod

#: Total number of arguments
_argc_numbers: int = None

#: Pointer to argv on the stack
_argv_ptr: int = None

#: Pointer to envp on the stack
_envp_ptr: int = None

#: Total number of environment variables
_envc_numbers: int = None

# Internal stack ptr
_stack_ptr: int = None
_was_updated = False


@pwndbg.dbg.event_handler(pwndbg.dbg_mod.EventType.START)
def update() -> None:
    if not pwndbg.dbg.selected_inferior().is_linux():
        return None

    global _stack_ptr
    global _was_updated
    # Captures the current stack pointer (SP) at the time of event START.
    # Note: This won't provide the SP value from the `_start` function
    # when attaching to an already running process, as `_start` has
    # already executed in that case.
    _stack_ptr = int(pwndbg.dbg.selected_frame().regs().by_name("sp"))
    _was_updated = False


def update_state() -> None:
    global _was_updated
    global _stack_ptr
    global _argc_numbers
    global _argv_ptr
    global _envp_ptr
    global _envc_numbers

    if _was_updated:
        return None
    _was_updated = True

    sp = _stack_ptr
    ptrsize = pwndbg.aglib.arch.ptrsize
    ptrbits = 8 * ptrsize

    try:
        _argc_numbers = pwndbg.aglib.memory.u(sp, ptrbits)
    except pwndbg.dbg_mod.Error:
        return None

    sp += ptrsize
    _argv_ptr = sp

    while pwndbg.aglib.memory.u(sp, ptrbits):
        sp += ptrsize

    sp += ptrsize
    _envp_ptr = sp

    _envc_numbers = 0
    try:
        while pwndbg.aglib.memory.u(sp, ptrbits):
            sp += ptrsize
            _envc_numbers += 1
    except pwndbg.dbg_mod.Error:
        pass


def argc() -> int:
    update_state()
    global _argc_numbers
    return _argc_numbers


def argv(number: int) -> pwndbg.dbg_mod.Value | None:
    update_state()

    global _argc_numbers
    global _argv_ptr

    if number > _argc_numbers:
        return None

    ppchar = pwndbg.aglib.typeinfo.pchar.pointer()
    argv = pwndbg.dbg.selected_inferior().create_value(_argv_ptr, ppchar)
    return (argv + number).dereference()


def envc() -> int:
    update_state()
    global _envc_numbers
    return _envc_numbers


def envp(number: int) -> pwndbg.dbg_mod.Value | None:
    update_state()

    global _envc_numbers
    global _envp_ptr

    if number > _envc_numbers:
        return None

    ppchar = pwndbg.aglib.typeinfo.pchar.pointer()
    envp = pwndbg.dbg.selected_inferior().create_value(_envp_ptr, ppchar)
    return (envp + number).dereference()


def environ(name: str) -> pwndbg.dbg_mod.Value | None:
    update_state()

    global _envc_numbers
    global _envp_ptr

    if not name:
        return None

    name += "="
    ppchar = pwndbg.aglib.typeinfo.pchar.pointer()
    envp = pwndbg.dbg.selected_inferior().create_value(_envp_ptr, ppchar)

    for i in range(_envc_numbers):
        ptr = (envp + i).dereference()
        sz = ptr.string()
        if sz.startswith(name):
            return ptr

    return None



from __future__ import annotations

from collections.abc import Iterator

import pwndbg
import pwndbg.aglib.memory
import pwndbg.aglib.typeinfo
import pwndbg.dbg_mod


def container_of(ptr: int, typename: str, fieldname: str) -> pwndbg.dbg_mod.Value:
    obj_addr = int(ptr) - pwndbg.aglib.typeinfo.load(typename).offsetof(fieldname)
    return pwndbg.aglib.memory.get_typed_pointer(typename, obj_addr)


def for_each_entry(
    head: pwndbg.dbg_mod.Value, typename: str, field: str
) -> Iterator[pwndbg.dbg_mod.Value]:
    head_addr = int(head.address)
    addr = head["next"]
    addr_int = int(addr)
    while addr_int != head_addr:
        yield container_of(addr_int, typename, field)
        addr = addr.dereference()["next"]
        addr_int = int(addr)


def swab(x: int) -> int:
    return int(
        ((x & 0x00000000000000FF) << 56)
        | ((x & 0x000000000000FF00) << 40)
        | ((x & 0x0000000000FF0000) << 24)
        | ((x & 0x00000000FF000000) << 8)
        | ((x & 0x000000FF00000000) >> 8)
        | ((x & 0x0000FF0000000000) >> 24)
        | ((x & 0x00FF000000000000) >> 40)
        | ((x & 0xFF00000000000000) >> 56)
    )


def _arr(x: pwndbg.dbg_mod.Value, n: int) -> pwndbg.dbg_mod.Value:
    """returns the nth element of type x, starting at address of x"""
    ptr = x.address.cast(x.type.pointer())
    return (ptr + n).dereference()


def compound_head(page: pwndbg.dbg_mod.Value) -> pwndbg.dbg_mod.Value:
    """returns the head page of compound pages"""
    assert page.type.code == pwndbg.dbg_mod.TypeCode.STRUCT and page.type.name_identifier == "page"
    # https://elixir.bootlin.com/linux/v6.2/source/include/linux/page-flags.h#L249
    head = page["compound_head"]
    if int(head) & 1:
        return (head - 1).cast(page.type.pointer()).dereference()

    pg_headty = pwndbg.aglib.typeinfo.load("enum pageflags")
    assert pg_headty is not None, "Type 'enum pageflags' not found"
    pg_head = pg_headty.enum_member("PG_head")
    assert pg_head is not None, "Type 'enum pageflags' not found"

    # https://elixir.bootlin.com/linux/v6.2/source/include/linux/page-flags.h#L212
    if int(page["flags"]) & (1 << pg_head):
        next_page = _arr(page, 1)

        head = next_page["compound_head"]
        if int(head) & 1:
            return (head - 1).cast(page.type.pointer()).dereference()

    return page



"""
Caches return values until some event in the inferior happens,
e.g. execution stops because of a SIGINT or breakpoint, or a
new library/objfile are loaded, etc.
"""

from __future__ import annotations

from collections import UserDict
from collections.abc import Callable
from enum import IntFlag
from functools import wraps
from typing import Any
from typing import TypeAlias
from typing import TypeVar

from typing_extensions import ParamSpec

T = TypeVar("T", covariant=True)
P = ParamSpec("P")

# Set to enable print logging of cache hits/misses/clears
NO_DEBUG, DEBUG_GET, DEBUG_CLEAR, DEBUG_SET = 0, 1, 2, 4
# combine the flags with | operator
debug = NO_DEBUG
# debug_name can be used to filter cache results by a given name
debug_name = "regs"


class DebugCacheDict(UserDict):  # type: ignore[type-arg]
    def __init__(self, func: Callable[P, T], *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.hits = 0
        self.misses = 0
        self.func = func
        self.name = f"{func.__module__.split('.')[-1]}.{func.__name__}"

    def __getitem__(self, key: tuple[Any, ...]) -> Any:
        if debug & DEBUG_GET and (not debug_name or debug_name in self.name):
            print(f"GET {self.name}: {key}")
        try:
            value = self.data[key]
            self.hits += 1
            return value
        except KeyError:
            self.misses += 1
            raise

    def __setitem__(self, key: tuple[Any, ...], value: Any) -> None:
        if debug & DEBUG_SET and (not debug_name or debug_name in self.name):
            print(f"SET {self.name}: {key}={value}")
        self.data[key] = value

    def clear(self) -> None:
        if debug & DEBUG_CLEAR and (not debug_name or debug_name in self.name):
            print(f"CLEAR {self.name} (hits: {self.hits}, misses: {self.misses})")
        self.data.clear()
        self.hits = 0
        self.misses = 0


Cache: TypeAlias = dict[tuple[Any, ...], Any] | DebugCacheDict


class _CacheUntilEvent:
    def __init__(self) -> None:
        self.caches: list[Cache] = []

    def connect_event_hooks(self, event_hooks: tuple[Any, ...]) -> None:
        """
        A given _CacheUntilEvent object may require multiple debugger events
        to be handled properly. E.g. our `stop` cache needs to be handled
        by `stop`, `mem_changed` and `reg_changed` events.
        """
        for event_hook in event_hooks:
            # This will just run a decorator (which will return a function, but
            # that return value is not used here).
            event_hook(self.clear)

    def clear(self) -> None:
        for cache in self.caches:
            cache.clear()

    def add_cache(self, cache: Cache) -> None:
        self.caches.append(cache)


# fmt: off
class CacheUntilEvent(IntFlag):
    """
    Not necessarily 1:1 with pwndbg.dbg_mod.EventTypes , but
    read the definition of that enum to get an idea of how these
    work.

    Notably, STOP is also triggered when the user changes debugee
    memory or register state.
    """
    STOP    = 0b00000001
    EXIT    = 0b00000010
    OBJFILE = 0b00000100
    START   = 0b00001000
    CONT    = 0b00010000
    THREAD  = 0b00100000
    FOREVER = 0b01000000

# fmt: on

# OR (|) events together to make a set
EventSet: TypeAlias = int


_ALL_CACHE_UNTIL_EVENTS: dict[CacheUntilEvent, _CacheUntilEvent] = {
    CacheUntilEvent.STOP: _CacheUntilEvent(),
    CacheUntilEvent.EXIT: _CacheUntilEvent(),
    CacheUntilEvent.OBJFILE: _CacheUntilEvent(),
    CacheUntilEvent.START: _CacheUntilEvent(),
    CacheUntilEvent.CONT: _CacheUntilEvent(),
    CacheUntilEvent.THREAD: _CacheUntilEvent(),
    CacheUntilEvent.FOREVER: _CacheUntilEvent(),
}

_NAME_TO_EVENT: dict[str, CacheUntilEvent] = {
    "stop": CacheUntilEvent.STOP,
    "exit": CacheUntilEvent.EXIT,
    "objfile": CacheUntilEvent.OBJFILE,
    "start": CacheUntilEvent.START,
    "cont": CacheUntilEvent.CONT,
    "thread": CacheUntilEvent.THREAD,
    "forever": CacheUntilEvent.FOREVER,
}
_ALL_CACHE_EVENT_NAMES = tuple(_NAME_TO_EVENT.keys())


def events_to_event_set(event_list: list[CacheUntilEvent]) -> EventSet:
    res = 0
    for an_event in event_list:
        res |= an_event.value
    return res


def connect_clear_caching_events(event_dicts: dict[CacheUntilEvent, tuple[Any, ...]]) -> None:
    """
    Connect given debugger event hooks to corresponding _CacheUntilEvent instances
    """
    for event_name, event_hooks in event_dicts.items():
        _ALL_CACHE_UNTIL_EVENTS[event_name].connect_event_hooks(event_hooks)


# A singleton used to mark a cache miss
_NOT_FOUND_IN_CACHE = object()
_KWARGS_SEPARATOR = object()

# Global value whether the results from cache are returned or not
# Used for debugging (by pwndbg.commands.memoize).
IS_CACHING = True


# Global value that allows disabling of individual cache types.
# This should only be set by the debugger at bring-up time. Thus
# it should be possible to perform this check at decoration time
# rather than at runtime and get a nice performance improvement, but
# I'm not sure how to do this safely exactly.
# The value is an event set (an OR (|) of different events)
IS_CACHING_DISABLED_FOR: EventSet = 0


def cache_until(*event_names: str) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """
    All possible values of the function arguments must be hashable
    (so e.g. `var: MyUnhashableType | None = None` is not allowed, but may fail rarely).
    The return value of the function should not be mutable, must not be mutated.
    """
    if any(event_name not in _ALL_CACHE_EVENT_NAMES for event_name in event_names):
        raise ValueError(
            f"Unknown event name[s] passed to the `cache_until` decorator: {event_names}.\n"
            f"Expected: {_ALL_CACHE_EVENT_NAMES}"
        )

    # We could require that CacheUntilEvent's be passed instead of strings as cache_until arguments.
    event_list: list[CacheUntilEvent] = [_NAME_TO_EVENT[event_name] for event_name in event_names]
    event_set: EventSet = events_to_event_set(event_list)

    def inner(func: Callable[P, T]) -> Callable[P, T]:
        if hasattr(func, "cache"):
            raise ValueError(
                f"Cannot cache the {func.__name__} function twice! "
                "Pass multiple event names to the `cache_until` decorator."
            )

        cache: Cache = {} if not debug else DebugCacheDict(func)

        @wraps(func)
        def decorator(*a: P.args, **kw: P.kwargs) -> T:
            if IS_CACHING and (event_set & IS_CACHING_DISABLED_FOR) == 0:
                key: tuple[Any, ...] = (a, _KWARGS_SEPARATOR, *kw.items())

                # Check if the value is in the cache; if we have a cache miss,
                # we return a special singleton object `_NOT_FOUND_IN_CACHE`. This way
                # we can also cache a result of 'None' from a function
                try:
                    cached_value: Any = cache.get(key, _NOT_FOUND_IN_CACHE)
                except TypeError:
                    print("Unhashable argument passed to a cache_until decorated function.")
                    print("Make the argument hashable or refactor.")
                    print(
                        f"Function: {func.__module__}.{func.__qualname__}",
                    )
                    print(f"Arguments: {repr(key)}")
                    assert False, "Unhashable argument passed to a cache_until decorated function."

                if cached_value is not _NOT_FOUND_IN_CACHE:
                    return cached_value

                value = func(*a, **kw)

                # Sanity check: its not perfect and won't cover all cases like ([],)
                # but it should be good enough
                if isinstance(value, list):
                    print(f"Should not cache mutable types! {func.__name__}")

                cache[key] = value

                return value

            return func(*a, **kw)

        # Set the cache on the function so it can be cleared on demand
        # this may be useful for tests
        decorator.cache = cache  # type: ignore[attr-defined]

        # Register the cache for the given event so it can be cleared
        for an_event in event_list:
            _ALL_CACHE_UNTIL_EVENTS[an_event].add_cache(cache)

        return decorator

    return inner


def clear_caches() -> None:
    for cache in _ALL_CACHE_UNTIL_EVENTS.values():
        cache.clear()


def clear_cache(cache_event: CacheUntilEvent) -> None:
    # I could imagine this being a hot path, so I don't want to do the
    # `str -> CacheUntilEvent` conversion here.
    _ALL_CACHE_UNTIL_EVENTS[cache_event].clear()


def clear_function_cache(func: Callable[..., T]) -> None:
    """
    Call this on a cached function/method to clear its cache. For methods,
    clears the cache for all object instances.
    """
    assert hasattr(func, "cache")
    getattr(func, "cache").clear()
